/*!
 * Copyright 2014 Smart Parking. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*!
 * @module smartcloud/BigQuery
 */

'use strict';

const { BigQuery: GCloudBigQuery } = require('@google-cloud/bigquery');

/**
 * SmartCloud BigQuery is a utility helper for interacting with BigQuery for cloud functions
 *
 * @constructor
 *
 * @example
 * const BigQuery = new BigQuery(options)
 *
 * @param {object} options - Configuration object containing the following properties:
 */
function BigQuery( options ) {
  if (!(this instanceof BigQuery)) {
    return new BigQuery(options);
  }
  this.options = options || {};
  this.bigquery;
}

/**
 * Returns registered BigQuery instance
 * We instantiate BigQuery on first call to prevent current auth error in cloud functions beta
 */
BigQuery.prototype.getUnderlyingBigQuery = function() {
  const self = this;
  if(!self.bigquery) {
    self.bigquery = new GCloudBigQuery();

    // TODO: watch out the following which comes out recently as of 12 Mar 2018 as the fix may evolve over time
    //     https://github.com/googleapis/nodejs-bigquery/issues/41
    self.bigquery.interceptors.push({
      request: reqOpts => {
        reqOpts.forever = true
        return reqOpts
      }
    });
  }
  return self.bigquery;
};

/**
 * Streams row(s) into a BigQuery table (template table)
 *
 * @param {object} rows - single object or array of objects to save
 * @param {object} options - Configuration object containing the following properties:
 * @param {string} datasetName - dataset name that target table belongs to
 * @param {string} tableName - name of the table to save data into
 * @param {string} templateSuffix - suffix to use for table instance
 *
 */
BigQuery.prototype.insertIntoBigQuery = function( rows, templateSuffix, options ) {
  const self = this;
  const dataset = options.datasetName;
  const table = options.tableName;

  const insertOptions = {
    templateSuffix: '_' + templateSuffix,
    ignoreUnknownValues: true
  }

  return self.getUnderlyingBigQuery().dataset(dataset).table(table).insert(rows, insertOptions)
  .catch(( err ) => {
    console.log(err);
    throw err;
    /*console.log('Error encountered while inserting rows into BigQuery.  Dataset or template table might not exist.  Make sure you run "npm run setup" in the function repo and are targeting the correct project in gcloud ("gcloud config list")');
    if(err.name === 'PartialFailureError') {
      err.errors.forEach(( err ) => {
        console.error('Encountered an error when inserting row: ', err.row)
        err.errors.forEach(( err ) => {
          console.error('--Reason: ', err.reason);
          console.error('--Message: ', err.message);
        })
      })
    } else if(err.response) {
      console.error('Insert errors:');
      err.response.insertErrors.forEach(( err ) => console.error(err));
    } else {
      throw err;
    }*/
  });
}

/**
 * Streams row(s) into a BigQuery table (Partitioned table)
 *
 * @param {object} rows - single object or array of objects to save
 * @param {object} options - Configuration object containing the following properties:
 * @param {string} datasetName - dataset name that target table belongs to
 * @param {string} tableName - name of the table to save data into
 *
 */
BigQuery.prototype.insertIntoPartitionedBigQuery = function( rows, options ) {
  const self = this;
  const dataset = options.datasetName;
  const table = options.tableName;

  const insertOptions = {
    ignoreUnknownValues: true
  }

  return self.getUnderlyingBigQuery().dataset(dataset).table(table).insert(rows, insertOptions)
  .catch(( err ) => {
    console.log(err);
    throw err;
  });
}

/**
 * Loads rows from a BigQuery table
 *
 * @param {string} query - query to run
 * @param {object} params - object containing named parameters and values
 *
 */
BigQuery.prototype.queryBigQuery = function( query, params ) {
  const self = this;
  if(params) {
    return self.getUnderlyingBigQuery().query({query: query, params: params});
  } else {
    return self.getUnderlyingBigQuery().query({query: query});
  }
}

module.exports = BigQuery;
