/*!
 * Copyright 2014 Smart Parking. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*!
 * @module smartcloud/BigTable
 */

'use strict';

const Result = require('./result.js');
const Errors = require('./errors.js');
const Transform = require('stream').Transform;
const util = require('util');
const OpenLocation = require('./openlocation.js')();
const { Bigtable: GCloudBigTable } = require('@google-cloud/bigtable');
const DEFAULT_INSTANCE = 'smart-cloud';

/**
 * SmartCloud BigTable is a utility helper for interacting with BigTable for cloud functions
 *
 * @constructor
 *
 * @example
 * const BigTable = new BigTable(options)
 *
 * @param {object} options - Configuration object containing the following properties:
 */
function BigTable( options ) {
  if (!(this instanceof BigTable)) {
    return new BigTable(options);
  }
  this.options = options || {};
  this.instance = this.options.bigTableInstance || DEFAULT_INSTANCE;
  this.bigtable;
}

/**
 * Returns registered BigTable module
 * We instantiate BigTable on first call to prevent current auth error in cloud functions beta
 */
BigTable.prototype.getUnderlyingBigTable = function() {
  const self = this;
  if(!self.bigtable) {
    self.bigtable = new GCloudBigTable();
  }
  return self.bigtable;
};

/**
 * Returns registered BigTable instance
 */
BigTable.prototype.getUnderlyingInstance = function() {
  const self = this;
  return self.getUnderlyingBigTable().instance(self.instance);
};

/**
 * Checks and logs partial errors.  Otherwise throw it.
 *
 * @param {object} errorObject - error returned by BigTable
 *
 */
BigTable.prototype.checkForPartialErrors = function( errorObject ) {
  const self = this;
  if(errorObject.name === 'PartialFailureError') {
    errorObject.errors.forEach(( error ) => {
      console.error(new Error(`Got a partial error from BigTable with code (${error.code}) and message: ${error.message}`))
      console.log('Row that caused error:', error.entry)
    })
  } else {
    if (errorObject instanceof Error) {
      console.error(errorObject);
    } else {
      console.error(new Error(errorObject));
    }
  }
}

/**
 * Insert a row into BigTable
 *
 * @param {object} row - row to insert into BigTable
 * @param {string} key - rowkey for neew row
 * @param {string} tableName - table to insert row into
 *
 */
BigTable.prototype.insertRowIntoBigTable = function( row, key, tableName ) {
  const self = this;
  const table = self.getUnderlyingInstance().table(tableName);
  const entries = [{ key: key, data: row }];
  return table.insert(entries).catch(( err ) => self.checkForPartialErrors(err))
}

/**
 * Inserts a batch of rows into BigTable
 *
 * @param {object} rows - an array of row objects (in format [{key:key, data:data}, {key:key, data:data}])
 * @param {string} tableName - table to insert rows into
 *
 */
BigTable.prototype.insertRowsIntoBigTable = function( rows, tableName ) {
  const self = this;
  const table = self.getUnderlyingInstance().table(tableName);
  return table.insert(rows).catch(( err ) => self.checkForPartialErrors(err));
}

/**
 * Query a small number of specific rows from BigTable
 * For larger queries, use streamRows functions instead
 *
 * @param {string...} rowkeys - array of row keys to get
 * @param {string} tableName - table to pull rows from
 * @param {obj} additionalOptions - additional query options (optional)
 *
 */
BigTable.prototype.getRows = function( rowkeys, tableName, additionalOptions ) {
  const self = this;
  const table = self.getUnderlyingInstance().table(tableName);
  // additionalOptions will override keys and filters if supplied
  let options = Object.assign({
    filter: [
      {
        column: {
          cellLimit: 1,
        },
      },
    ],
  }, additionalOptions);
  if (rowkeys) {
    options.keys = rowkeys;
  }

  return table.getRows(options)
  .then(( data ) => {
    const rows = data[0];
    const results = [];
    rows.forEach(( row ) => {
      results.push(new Result(row))
    })
    return Promise.resolve(results);
  })
  .catch(( err ) => self.checkForPartialErrors(err));
}

/**
 * Query single row from BigTable
 *
 * @param {string} rowkey - key of row to get
 * @param {string} tableName - table to pull row from
 * @param {obj} options - additional query options (optional)
 *
 */
BigTable.prototype.getRow = function( rowkey, tableName, additionalOptions ) {
 const self = this;
 const table = self.getUnderlyingInstance().table(tableName);
 const row = table.row(rowkey);
 return row.get(additionalOptions)
 .then(( [row] ) => {
   const result = new Result(row);
   result.data = row;
   result.id = rowkey;
   return Promise.resolve(result)
 })
 .catch(( error ) => {
   console.log("Encountered an error while trying to load row from BigTable: ", error);
   return Promise.resolve(new Result())
 })
}

/**
 * Get a stream of rows from BigTable
 *
 * @param {string} tableName - table to pull rows from
 * @param {obj} options - query options
 *
 */
BigTable.prototype.streamRows = function( tableName, options ) {
  const self = this;
  const baseOptions = {
    filter: [
      {
        column: {
          cellLimit: 1,
        },
      },
    ],
  };
  const table = self.getUnderlyingInstance().table(tableName);
  return table.createReadStream(Object.assign(baseOptions, options));
}

/**
 * Convert BigTable stream into a different object format
 *
 * @param targetType - (result || object) the object type we want output from the transform
 */
function ConvertStream( targetType, options ) {
  if (!(this instanceof ConvertStream)) {
    return new ConvertStream(targetType, options);
  }

  if (!options) options = {};
  options.objectMode = true;
  Transform.call(this, options);
  this.targetType = targetType;
}
util.inherits(ConvertStream, Transform);

ConvertStream.prototype._transform = function( row, encoding, callback ) {
  const self = this;
  var convertedObject;
  switch(self.targetType.toUpperCase()) {
    case "OBJECT": convertedObject = new Result(row).getJson(); break;
    default: convertedObject = new Result(row); break;
  }
  this.push(convertedObject);
  callback();
};

/**
 * Get a stream of Results from BigTable
 *
 * @param {string} tableName - table to pull rows from
 * @param {obj} options - query options
 *
 */
BigTable.prototype.streamResults = function( tableName, options ) {
  const self = this;
  const baseOptions = {
    filter: [
      {
        column: {
          cellLimit: 1,
        },
      },
    ],
  };
  const table = self.getUnderlyingInstance().table(tableName);
  const transform = new ConvertStream('result');
  table.createReadStream(Object.assign(baseOptions, options))
    .on('error', ( error ) => { console.error(error); transform.emit('error', error) })
    .on('data', ( row ) => { transform.write(row) })
    .on('end', ( ) => { transform.end() })
  return transform;
}

/**
 * Get a stream of decoded objects from BigTable
 *
 * @param {string} tableName - table to pull rows from
 * @param {obj} options - query options
 *
 */
BigTable.prototype.streamObjects = function( tableName, options ) {
  const self = this;
  const baseOptions = {
    filter: [
      {
        column: {
          cellLimit: 1,
        },
      },
    ],
  };
  const table = self.getUnderlyingInstance().table(tableName);
  const transform = new ConvertStream('object');
  table.createReadStream(Object.assign(baseOptions, options))
    .on('error', ( error ) => { console.error(error); transform.emit('error', error) })
    .on('data', ( row ) => { transform.write(row) })
    .on('end', ( ) => { transform.end() })
  return transform;
}

/**
 * Get a stream of rows from BigTable that fall in a range
 *
 * @param {string} tableName - table to pull rows from
 * @param {string} startRowKey - rowkey to start range from
 * @param {string} endRowKey - rowkey to end range from
 * @param {obj} options - additional query options (optional)
 *
 */
BigTable.prototype.streamRowsInRange = function( tableName, startRowKey, endRowKey, options ) {
  const self = this;
  return self.streamRows(tableName, Object.assign({
    start: startRowKey,
    end: endRowKey,
  }, options));
}

/**
 * Converts a location object to BigTable friendly fields (i.e. decimals must be strings)
 *
 * @param {obj} location - the object containing location information
 *
 */
BigTable.prototype.scrubLocation = function( location ) {
  if(!location) return;
  if(location.latitude && location.longitude) {
    location.latitude = `${location.latitude}`;
    location.longitude = `${location.longitude}`;
    location.altitude = location.altitude ? `${location.altitude}` : '180';
    location.openLocationCode =  OpenLocation.encode(location.latitude, location.longitude)
    return location;
  } else if(location.openLocationCode) {
    const codeArea = OpenLocation.decode(location.openLocationCode);
    location.latitude = `${codeArea.latitudeCenter}`;
    location.longitude = `${codeArea.longitudeCenter}`;
    location.altitude = '180';
    return location;
  } else {
    throw new Errors.NoRetryError('Could not convert location definition to BigTable format.  Processing aborted.');
  }
}

/**
 * Filter single row from BigTable, mutate the row under certain conditions 
 * - row.js shows that row.filter() calls internal api of checkAndMutate()
 * 
 * @param {string} tableName - table to pull row from
 * @param {string} rowKey - the row key for this row
 * @param {obj} filter - the filter object that performs column/family/value checks
 * @param {obj} config - the result of the filter perforn on the field of onMatch and onNoMatch fields
 */
BigTable.prototype.mutateRowWithFilter = function( tableName, rowKey, filter, config ) {
  const self = this;
  const table = self.getUnderlyingInstance().table(tableName);
  const row = table.row(rowKey);

  return row.filter(filter, config)
  .then(( data ) => {
    return Promise.resolve(data)
  })    
  .catch((err) => {
    console.log('row.filter error: ' + err);
    return Promise.reject(err);
  })
}

/**
 * Delete single row from BigTable
 *
 * @param {string} rowkey - key of row to delete
 * @param {string} tableName - table to remove row from
 *
 */
BigTable.prototype.deleteRow = function( rowkey, tableName ) {
 const self = this;
 const table = self.getUnderlyingInstance().table(tableName);
 const row = table.row(rowkey);
 return row.delete();
}

module.exports = BigTable;
