/*!
 * Copyright 2014 Smart Parking. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*!
 * @module smartcloud/Common
 */

'use strict';

// dependencies
const uuid = require('uuid');
const crypto = require('crypto');
const Errors = require('./errors.js');

/**
 * SmartCloud is a utility helper for interacting with the Smartcloud and GCP platforms
 *
 * @constructor
 *
 * @example
 * const Common = new Common(options)
 *
 * @param {object} options - Configuration object containing the following properties:
 */
function Common( options ) {
  if (!(this instanceof Common)) {
    return new Common(options);
  }
  this.options = options;
}

/**
 * Returns a unique ID to be used within the SmartCloud platform
 *
 */
Common.prototype.getUUID = function() {
  return uuid.v4();
};

/**
 * Returns a datastore / bigtable friendly hash of a string
 *
 */
Common.prototype.hash = function(stringToHash) {
  return crypto.createHash('md5').update(stringToHash).digest('hex');
};

/**
 * Returns the o/r/s/s/u reference for a given object or undefined if
 * the site structure reference cannot be found within the given obj parameter.
 *
 * @param {object} obj - target object
 *
 */
Common.prototype.getReference = function(obj) {
  if (typeof obj === 'undefined') {
    throw new Errors.NoRetryError('getReference function cannot handle the request as parameter object is undefined.')
  }

  let ref;

  if (obj.hasOwnProperty('ref')) {
    // object destructuring for values in obj.ref
    const { org = undefined,
            region = undefined,
            site = undefined,
            sector = undefined,
            unit = undefined
          } = obj.ref;

    if (org) ref = org;
    if (region) ref += `.${region}`;
    if (site) ref += `.${site}`;
    if (sector) ref += `.${sector}`;
    if (unit) ref += `.${unit}`;
  } else {
    // object destructuring for values in obj
    const { orgId = undefined,
            regionId = undefined,
            siteId = undefined,
            sectorId = undefined,
            unitId = undefined
          } = obj;

    if (orgId) ref = orgId;
    if (regionId) ref += `.${regionId}`;
    if (siteId) ref += `.${siteId}`;
    if (sectorId) ref += `.${sectorId}`;
    if (unitId) ref += `.${unitId}`;
  }

  return ref;
};

module.exports = Common;
