/*!
 * Copyright 2014 Smart Parking. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*!
 * @module smartcloud/Config
 */

'use strict';

const BASE_ENV_PREFIX = 'SMARTCLOUD';

/**
 * SmartCloud Config is a utility helper for interacting with runtime configs for cloud functions
 *
 * @constructor
 *
 * @example
 * const Config = new Config(options)
 *
 * @param {object} options - Configuration object containing the following properties:
 */
function Config( options ) {
  if (!(this instanceof Config)) {
    return new Config(options);
  }
  this.options = options || {};
  this.config;
}

/**
 * Returns registered config instance
 */
Config.prototype.getUnderlyingConfig = function() {
  const self = this;
  if(!self.config) {
    self.config = require('googleapis');
  }
  return self.config;
};

/**
 * @desc Gets variable from environment
 *
 * @param {string} variableName The variable name of the variable to return
 */
Config.prototype.getEnvVariable = function( configName, variableName ) {
  const self = this;
  const envVariable = `${BASE_ENV_PREFIX}_${configName.toUpperCase()}_${variableName.toUpperCase()}`
  const value = process.env[envVariable];
  if(!value) return undefined;
  if(value == 'undefined') return undefined;
  return value;
};

/**
 * runtimeConfig.getVariable
 *
 * @desc Reads a runtime config value
 *
 * @param {string} configName The config name of the variable to return
 * @param {string} variableName The variable name of the variable to return
 * @return {Promise} Promise that resolves the variable value
 */
Config.prototype.getRuntimeConfigVariable = function(configName, variableName) {
  const self = this;

  const envVariable = self.getEnvVariable(configName, variableName);
  if(envVariable) return Promise.resolve(envVariable);

  const runtimeConfig = self.getUnderlyingConfig().runtimeconfig('v1beta1');
  return self.auth()
  .then(( authClient ) => {
    const projectId = process.env.GCP_PROJECT;
    const fullyQualifiedName = `projects/${projectId}/configs/${configName}/variables/${variableName}`;
    return new Promise(function( resolve, reject ) {
      runtimeConfig.projects.configs.variables.get({
         auth: authClient,
         name: fullyQualifiedName,
      }, function( err, variable ) {
         if (err) {
           reject(err);
         } else {
           resolve(variable.text);
         }
      });
    });
  });
}

/**
 * auth
 *
 * @desc Authenticates using default credentials
 *
 * @return {Promise} Promise that resolves an authClient
 */
Config.prototype.auth = function() {
  const self = this;
  return new Promise(function( resolve, reject ) {
    self.getUnderlyingConfig().auth.getApplicationDefault(function( err, authClient, projectId ) {
      if(err) {
        reject(err);
        return;
      }

      if(authClient.createScopedRequired && authClient.createScopedRequired()) {
        const scopes = [
          'https://www.googleapis.com/auth/cloud-platform',
          'https://www.googleapis.com/auth/cloudruntimeconfig'
        ];
        authClient = authClient.createScoped(scopes);
      }

      resolve(authClient);
    });
  });
}

module.exports = Config;
