/*!
 * Copyright 2014 Smart Parking. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*!
 * @module smartcloud/Datastore
 */

'use strict';

const { Datastore: GCloudDatastore } = require('@google-cloud/datastore');

/**
 * SmartCloud Datastore is a utility helper for interacting with Datastore for cloud functions
 *
 * @constructor
 *
 * @example
 * const Datastore = new Datastore(options)
 *
 * @param {object} options - Configuration object containing the following properties:
 * @param {string} options.maxRetries - Number of maxRetries
 * @param {string} options.aggregateType - Aggregate type for loading existing records
 * @param {array} options.indexColumns - column names to index when storing in datastore
 */
function Datastore( options ) {
  if (!(this instanceof Datastore)) {
    return new Datastore(options);
  }
  this.options = options || {};

  this.indexes = this.options.indexColumns
                 ? this.options.indexColumns.map((value) => value.toUpperCase().trim())
                 : [];

  this.maxRetries = this.options.maxRetries || 10;
  this.datastore;
}

/**
 * Returns registered datastore instance
 * We instantiate datastore on first call to prevent current auth error in cloud functions beta
 */
Datastore.prototype.getUnderlyingDatastore = function() {
  const self = this;
  if(!self.datastore) {
    self.datastore = new GCloudDatastore({ maxRetries: self.maxRetries });
  }
  return self.datastore;
};

/**
 * Convert a standard JS object to datastore array with each field marked to be
 * exlcuded from indexing (i.e., remove all auto-generated indexing)
 *
 * @param {object} obj - [Required] Object to be converted.
 *
 */
Datastore.prototype.convertObjectToUnindexedData = function( obj ) {
  const self = this;
  const data = [];
  Object.keys(obj).forEach(( key ) => {
    const nestedObject = obj[key];
    if(nestedObject instanceof Array) {
      data.push({name: key, value: JSON.stringify(nestedObject), excludeFromIndexes: true});
    } else {
      data.push({
        name: key,
        value: nestedObject,
        excludeFromIndexes: !(self.indexes.indexOf(key.toUpperCase().trim()) > -1)
      });
    }
  });
  return data;
}

/**
 * Save object to datastore
 *
 * @param {object} entityKey - Datastore Key to use for identifying saved entity.
 * @param {object} entity - Object to save.
 *
 */
Datastore.prototype.saveToDatastore = function( entity, entityKey ) {
  const self = this;
  return self.getUnderlyingDatastore().save({
    key: entityKey,
    data: self.convertObjectToUnindexedData(entity)
  });
}

/**
 * Run a query againt datastore.  Returns only entities as array
 *
 * @param {object} query - Datastore query object to run
 *
 */
Datastore.prototype.queryDatastore = function( query ) {
  const self = this;
  return self.getUnderlyingDatastore().runQuery(query)
  .then(( results ) => {
    if(!results) return Promise.resolve([]);
    var entities = results[0];
    if(!(entities instanceof Array)) entities = [entities];
    return Promise.resolve(entities);
  });
}

/**
 * Get object from datastore
 *
 * @param {object} entityKey - Datastore Key to use for identifying saved entity.
 *
 */
Datastore.prototype.loadFromDatastore = function( entityKey ) {
  const self = this;
  return self.getUnderlyingDatastore().get(entityKey);
}

/**
 * Delete object from datastore
 *
 * @param {object} entityKey - Datastore key to use for identifying the saved entity or alternatively an array of datastore keys.
 *
 */
Datastore.prototype.deleteFromDatastore = function( entityKey ) {
  const self = this;
  return self.getUnderlyingDatastore().delete(entityKey);
}

/**
 * Returns key for datastore.  Same as calling Datastore().key(options)
 *
 * @param {object} options - [Required] Options for defining key
 * @param {...*} options.path - Key path.
 * @param {string=} options.namespace - Optional namespace.
 *
 */
Datastore.prototype.datastoreKey = function( options ) {
  const self = this;
  return self.getUnderlyingDatastore().key( options );
}

/**
 * Save record to datastore
 *
 * @param {object} entity - Object to save.
 * @param {...*} entityPath - ID for object to save or explicity array path for entity
 *
 * @example
 * saveRecord(obj, 'badger.123').then();
 * saveRecord(obj, ['org', 'acme', 'site', 'rocketMall']).then();
 */
Datastore.prototype.saveRecord = function( entity, path ) {
  const self = this;
  if(!(path instanceof Array)) {
    path = [self.options.aggregateType, path]
  }
  const entityKey = self.datastoreKey(path);
  return self.saveToDatastore(entity, entityKey);
}

/**
 * Save multiple records to datastore
 *
 * @param [objects] entity - Objects to save.
 *
 * @example
 * saveRecord([
 *   {key: datastoreKey1, data:{'foo':'bar'}},
 *   {key: datastoreKey2, data:{'rocket':'raccoon'}}
 * ]).then();
 */
Datastore.prototype.saveRecords = function( entities ) {
  const self = this;
  for(var i = 0; i < entities.length ; i++) {
    entities[i].data = self.convertObjectToUnindexedData(entities[i].data)
  }
  return self.getUnderlyingDatastore().save(entities);
}

/**
 * Creates and returns a base datastore query object
 *
 * @param {string} entityType - name of the target table.
 *
 */
Datastore.prototype.createDatastoreQuery = function( entityType ) {
  const self = this;
  return self.getUnderlyingDatastore().createQuery(entityType);
}

module.exports = Datastore;
