/*!
 * Copyright 2014 Smart Parking. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*!
 * @module smartcloud/Deploy
 */
'use strict';

const TEST_ENV_VARIABLE_TO_CHECK = 'SMARTCLOUD_PUBSUB_ROUTER_PORT';
const FUNCTIONS_EMULATOR_NODE_VERSION = 'v12.18.2';

/**
 * Deploy is a helper for deploying cloud functions
 *
 * @constructor
 *
 * @example
 * const Deploy = new Deploy(options)
 *
 * @param {object} options - Configuration object
 */
function Deploy(app, options) {
  if (!(this instanceof Deploy)) {
    return new Deploy(app, options);
  }
  this.functionName = 'Unknown';
  this.functionDetails = {};
  this.app = app;
  this.options = options;
  this.testEnvironment = false;
  this.shellService;
  this.nodeVersion;
}

/**
 * Manually set function being handled by Deploy
 */
Deploy.prototype.setApp = function (app) {
  this.app = app;
}

/**
 * Get command to set node version in sub shell
 */
Deploy.prototype.precommand = function () {
  return `. ~/.nvm/nvm.sh; nvm use ${this.nodeVersion}; `;
}

/**
 * Returns true if a testing environment has been detected
 */
Deploy.prototype.isTestingEnvironment = function () {
  return this.testEnvironment;
}

/**
 * Returns shell inferface
 * @private
 */
Deploy.prototype.shell = function () {
  const self = this;
  if (!self.shellService) self.shellService = require('shelljs');
  return self.shellService;
}

/**
 * Creates formatted log message
 * @private
 */
Deploy.prototype.log = function (message) {
  const self = this;
  console.log(`[${self.functionDetails.functionName}]  `, message)
}

/**
 * Gets variable from env.  Return null if not found.
 */
Deploy.prototype.getEnvVariable = function (variableName) {
  const self = this;
  const value = process.env[variableName];
  if (!value) return undefined;
  if (value == 'undefined') return undefined;
  return value;
};

/**
 * Checks if environment has needed service to perform Deploy
 */
Deploy.prototype.checkEnvironment = function () {
  const self = this;
  return self.checkForTestEnvironment()
    .then((results) => { return self.loadAppDetails() })
    .then((results) => { return self.checkForRequiredFunctions() })
    .then((results) => { return self.checkNodeVersion() })
    .then((results) => { return self.startLocalEmulator() })
    .then((results) => { return self.loadProjectId() })
};

/**
 * Check if local env variable indicates a testing environment is in use
 * @private
 */
Deploy.prototype.checkForTestEnvironment = function () {
  const self = this;
  self.testEnvironment = self.getEnvVariable(TEST_ENV_VARIABLE_TO_CHECK) ? true : false;
  if (self.testEnvironment) {
    console.log("[Testing settings detected in env variables.  Deployment will target local emulators.]")
  }
  return Promise.resolve();
}

/**
 * Load details from the associated function
 * @private
 */
Deploy.prototype.loadAppDetails = function () {
  const self = this;
  if (!self.app) throw new Error('No app supplied in deployment options, cannot deploy.');
  if (self.app.hasOwnProperty('getFunctionDetails')) {
    self.functionDetails = self.app.getFunctionDetails();
  } else {
    throw new Error('app does not export getFunctionDetails function.  Please add to continue with deployment.');
  }
  return Promise.resolve();
}

/**
 * Checks if required command-line functions are installed
 * @private
 */
Deploy.prototype.checkForRequiredFunctions = function () {
  const self = this;
  return Promise.all([
    self.checkForGcloudCli(),
    //self.checkForFunctionsCli(),
    self.checkForNVM()
  ])
}

/**
 * Checks if we have glcoud installed
 * @private
 */
Deploy.prototype.checkForGcloudCli = function () {
  const self = this;
  if (self.shell().which('gcloud')) {
    return Promise.resolve();
  } else {
    return Promise.reject('glcoud cli is required to run setup.  Refer to: https://cloud.google.com/sdk/downloads');
  }
}

/**
 * Checks if we have functions emulator installed
 * @private
 */
Deploy.prototype.checkForFunctionsCli = function () {
  const self = this;
  if (self.shell().which('functions')) {
    return Promise.resolve();
  } else {
    return Promise.reject('functions cli is required to deploy to local testing environment.  Refer to: https://cloud.google.com/functions/docs/emulator');
  }
}

/**
 * Checks if we have NVM installed
 * @private
 */
Deploy.prototype.checkForNVM = function () {
  const self = this;
  return new Promise((resolve, reject) => {
    if (self.shell().exec('. ~/.nvm/nvm.sh; nvm version', { silent: true }).code !== 0) {
      reject('nvm is required to deploy to local testing environment.  Refer to: https://github.com/creationix/nvm/blob/master/README.md');
    } else {
      self.nodeVersion = self.shell().exec('. ~/.nvm/nvm.sh; nvm version', { silent: true }).stdout.trim();
      resolve();
    }
  })
}

/**
 * Checks for current project name
 */
Deploy.prototype.loadProjectId = function () {
  const self = this;
  return new Promise((resolve, reject) => {
    if (self.shell().exec('gcloud config get-value project 2>/dev/null', { silent: true }).code !== 0) {
      reject('cannot get gcloud project');
    } else {
      self.projectId = self.shell().exec('gcloud config get-value project 2>/dev/null', { silent: true }).stdout.trim();
      resolve();
    }
  });
}

/**
 * Checks for required node version for local functions emulator
 * @private
 */
Deploy.prototype.checkNodeVersion = function () {
  const self = this;
  if (!self.testEnvironment) return Promise.resolve();

  if (self.isNodeVersionInstalled(FUNCTIONS_EMULATOR_NODE_VERSION)) {
    self.checkNodeVersionRequirements();
    return self.reinstallModules(FUNCTIONS_EMULATOR_NODE_VERSION);
  } else {
    return self.installEmulatorNode();
  }
}

/**
* Checks if a version of node is installed
* @private
 */
Deploy.prototype.isNodeVersionInstalled = function (nodeVersion) {
  const self = this;
  return self.shell().exec(`. ~/.nvm/nvm.sh; nvm use ${nodeVersion}`, { silent: true }).code == 0;
}


/**
 * Prints warnings if our default nvm differs from the one required by the emulator
 */
Deploy.prototype.checkNodeVersionRequirements = function () {
  const self = this;
  if (self.nodeVersion.trim() !== FUNCTIONS_EMULATOR_NODE_VERSION.trim()) {
    self.log('WARNING: Default node version is different from version required for local emulator')
    self.log(`WARNING: Run 'nvm alias default ${FUNCTIONS_EMULATOR_NODE_VERSION}' to make deployments faster`)
  }
}

/**
* Installs version of node requried for cloud emulator
* @private
*/
Deploy.prototype.installEmulatorNode = function () {
  const self = this;
  return self.installNodeVersion(FUNCTIONS_EMULATOR_NODE_VERSION)
    .then((results) => { self.reinstallModules(FUNCTIONS_EMULATOR_NODE_VERSION) })
    .then((results) => { self.installEmulator() });
}

/**
 * Installed the version of node required
 * @private
 */
Deploy.prototype.installNodeVersion = function (nodeVersion) {
  const self = this;
  self.log(`Installing node ${nodeVersion}`);
  return new Promise((resolve, reject) => {
    if (self.shell().exec(`. ~/.nvm/nvm.sh; nvm install ${FUNCTIONS_EMULATOR_NODE_VERSION}`).code !== 0) {
      reject('Error installing required node version for local functions emulator.  Deployment will stop');
    } else {
      resolve();
    }
  })
}

/**
 * Install emulator on require version
 * @private
 */
Deploy.prototype.installEmulator = function () {
  const self = this;
  self.log(`Installing cloud function emulator`);
  return new Promise((resolve, reject) => {
    self.shell().exec(`${self.precommand()} functions stop`, { silent: true });
    self.shell().exec(`${self.precommand()} functions kill`, { silent: true });
    self.shell().exec(`${self.precommand()} npm uninstall -g @google-cloud/functions-emulator`, { silent: true });
    if (self.shell().exec(`${self.precommand()} npm install -g @google-cloud/functions-emulator --loglevel=error`).code !== 0) {
      reject('Error installing functions emulator.  Deployment will stop');
    } else {
      resolve();
    }
  })
}

/**
 * Starts local emulator if we're in a testing environment
 * @private
 */
Deploy.prototype.startLocalEmulator = function () {
  const self = this;
  return new Promise((resolve, reject) => {
    if (self.testEnvironment) {
      if (self.shell().exec(`${self.precommand()} functions start`, { silent: true }).code !== 0) {
        reject('could not start local cloud function emulator.  Deployment will stop.');
      } else {
        resolve();
      }
    } else {
      resolve();
    }
  })
}

/**
 * Run tests for app
 */
Deploy.prototype.runTests = function () {
  const self = this;
  self.log('running tests')
  return new Promise((resolve, reject) => {
    if (self.shell().exec(`${self.precommand()} npm test`, { silent: true }).code !== 0) {
      reject(`tests failed.  Run 'npm test' in folder to see details.  Deployment will be stopped`);
    } else {
      resolve();
    }
  })
}

/**
 * Run setup for app
 */
Deploy.prototype.runSetup = function () {
  const self = this;
  self.log('running setup')
  return new Promise((resolve, reject) => {
    if (self.shell().exec(`${self.precommand()} node setup`).code !== 0) {
      reject(`setup failed.  Run 'node setup' in folder to see details.  Deployment will be stopped`);
    } else {
      resolve();
    }
  })
}

/**
 * Run install for app
 */
Deploy.prototype.reinstallModules = function (nodeVersion) {
  const self = this;
  if (self.nodeVersion.trim() === nodeVersion.trim()) return Promise.resolve();
  self.nodeVersion = nodeVersion;
  self.log(`installing required node modules for ${nodeVersion} (this may take awhile)...`)
  return new Promise((resolve, reject) => {
    self.shell().exec('rm -rf node_modules', { silent: true });
    if (self.shell().exec(`${self.precommand()} npm install --loglevel=error`, { silent: true }).code !== 0) {
      reject(`install failed.  Run 'npm install' in folder to see details.  Deployment will be stopped`);
    } else {
      resolve();
    }
  })
}

/**
 * Get storage bucket info from function details
 * @private
 */
Deploy.prototype.getTriggerNameParameter = function () {
  const self = this;
  return self.functionDetails.trigger || '';
}

/**
 * Get storage bucket info from function details
 * @private
 */
Deploy.prototype.getTriggerResourceParameter = function () {
  const self = this;
  return self.functionDetails.trigger || '';
}

/**
 * Get flags for gcloud deployment command.
 * @private
 */
Deploy.prototype.getDeployFlags = function () {
  const self = this;
  let flags = '';

  if (self.functionDetails.memory) {
    flags += '--memory=' + self.functionDetails.memory;
  } else {
    flags += '--memory=128MB';
  }

  if (self.functionDetails.timeout) {
    flags += ' --timeout=' + self.functionDetails.timeout;
  }

  if (self.functionDetails.triggerType === '--trigger-http') {
    return flags;
  }

  if (self.functionDetails.retry) {
    flags += ' --retry';
  }

  return flags;
}

/**
 * Deploy to local function emulator
 * @private
 */
Deploy.prototype.deployLocal = function () {
  const self = this;
  self.log('deploying locally');
  const functionName = self.functionDetails.functionName;
  const triggerResource = self.getTriggerResourceParameter();

  const triggerParamName = self.functionDetails.triggerType.toLowerCase() === 'pubsub' ? '--trigger-topic' : '--trigger-bucket';

  const deployCommand = `functions deploy ${functionName} ${triggerParamName} ${triggerResource}`
  return new Promise((resolve, reject) => {
    if (self.shell().exec(self.precommand() + deployCommand, { silent: true }).code !== 0) {
      reject("Deployment failed");
    } else {
      resolve();
    }
  })
}

/**
 * Deploy function to GCP
 * @private
 */
Deploy.prototype.deployToGCP = function () {
  const self = this;
  self.log(`deploying to GCP: ${self.projectId}`);

  let deployCommand = '';

  const functionName = self.functionDetails.functionName;

  if (self.functionDetails.triggerType && self.functionDetails.triggerType.toLowerCase() === 'http') {
    deployCommand = `gcloud functions deploy ${functionName} --trigger-http`;
  } else {
    const triggerResource = self.getTriggerResourceParameter();
    const deployFlags = self.getDeployFlags();
    const triggerParamName = self.functionDetails.triggerType.toLowerCase() === 'pubsub' ? '--trigger-topic' : '--trigger-bucket';
    deployCommand = `gcloud functions deploy ${functionName} ${triggerParamName} ${triggerResource} ${deployFlags}`;
  }

  if (self.functionDetails.vpcConnector) {
    deployCommand += ` --vpc-connector projects/${self.projectId}${self.functionDetails.vpcConnector}`;
  }

  //deployCommand = `${deployCommand} --runtime ${(self.functionDetails.node10 === true ? 'nodejs10' : 'nodejs8')}`;
  deployCommand = `${deployCommand} --runtime nodejs12 --set-env-vars GCP_PROJECT=${self.projectId}`;
  console.log(deployCommand);

  return new Promise((resolve, reject) => {
    if (self.shell().exec(deployCommand).code !== 0) {
      reject("Deployment failed");
    } else {
      resolve();
    }
  })
}


/**
 * Run deploy for app
 * @private
 */
Deploy.prototype.runDeploy = function () {
  const self = this;
  if (self.testEnvironment) {
    return self.deployLocal();
  } else {
    return self.deployToGCP();
  }
}

/**
 * Run deployment for app
 */
Deploy.prototype.deploy = function () {
  const self = this;
  return self.checkEnvironment()
    .then((results) => { return self.runTests() })
    .then((results) => { return self.runSetup() })
    .then((results) => { return self.runDeploy() })
    .then(() => { self.log('deployment successful') })
    .catch((error) => {
      self.log(error);
      throw error;
    });
}

module.exports = Deploy;
