/*!
 * Copyright 2014 Smart Parking. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*!
 * @module smartcloud/Memcache
 */

'use strict';

const DatastoreModule = require('./datastore.js');
const RuntimeConfigModule = require('./config.js');

/**
 * SmartCloud Memcache is a utility helper for interacting with Memcache for cloud functions
 *
 * @constructor
 *
 * @example
 * const Memcached = new Memcached(options)
 *
 * @param {object} options - Configuration object containing the following properties:
 * @param {string} options.maxRetries - Number of maxRetries
 */
function Memcached( options ) {
  if (!(this instanceof Memcached)) {
    return new Memcached(options);
  }
  this.options = options || {};
  this.maxRetries = this.options.maxRetries || 10;
  this.memcache;
  this.datastore = DatastoreModule(options);
  this.config = RuntimeConfigModule(options);
}

/**
 * Returns registered memcache instance
 * We instantiate memcache on first call to prevent current auth error in cloud functions beta
 */
Memcached.prototype.getUnderlyingMemcache = function() {
  const self = this;
  if(self.Memcached === undefined) {
    return Promise.all([
      self.config.getRuntimeConfigVariable('memcached', 'host'),
      self.config.getRuntimeConfigVariable('memcached', 'port')
    ])
    .then(( [host, port] ) => {
      var memcachedURL = host + ':' + port;
      const MemcachedModule = require('memcached');
      self.memcache = new MemcachedModule(memcachedURL);
      return new Promise(( resolve, reject ) => {
        self.memcache.connect(memcachedURL, function( err, conn ) {
          if(err) {
            return reject(err);
          } else {
            return resolve(self.memcache);
          }
        });
      });
    });
  } else {
    return Promise.resolve(self.memcache);
  }
};

Memcached.prototype.saveValueToDatastore = function( key, value ) {
  const self = this;
  const entityKey = self.datastore.datastoreKey(['Memcached', key]);
  const data = { value: value };
  return self.datastore.saveToDatastore(data, entityKey);
}

Memcached.prototype.getValueFromDatastore = function( key ) {
  const self = this;
  const entityKey = self.datastore.datastoreKey(['Memcached', key]);
  return self.datastore.loadFromDatastore(entityKey)
  .then(( [record] ) => {
    const value = record.value;
    self.memcachedSet(key, value);
    return Promise.resolve(value);
  })
}

/**
 * Fetches the value associated with the key from memcached, store value to memcached if not exists
 *
 * @param {string} key - The key for the value to retrieve.
 *
 */
Memcached.prototype.getValueFromMemcached = function( key ) {
  const self = this;
  var source;

  return self.getUnderlyingMemcache()
  .then(( connection ) => {
    return self.memcachedGet(connection, key);
  })
  .then(( result ) => {
    if(result) {
      source = 'memcached'
      return Promise.resolve(result);
    } else {
      source = 'datastore'
      return self.getValueFromDatastore(key);
    }
  })
  .then(( value ) => {
    return Promise.resolve({
      value: value,
      source: source
    })
  })
}

/**
 * Fetches the value associated with the key from memcached
 *
 * @param {string} key - The key for the value to retrieve.
 *
 */
Memcached.prototype.memcachedGet = function( key ) {
  const self = this;
  return self.getUnderlyingMemcache()
  .then(( connection ) => {
    return new Promise(( resolve, reject ) => {
      connection.get(key, function( err, data ) {
        if(err) {
          return reject(err);
        } else {
          return resolve(data);
        }
      });
    })
  });
}

/**
 * Sets the value associated with the key in memcache
 *
 * @param {string} key - the key of the key/value pair.
 * @param {string} value - The value of the key/value pair.
 *
 */
Memcached.prototype.memcachedSet = function( key, value ) {
  const self = this;
  return self.getUnderlyingMemcache()
  .then(( connection ) => {
    return new Promise(( resolve, reject ) => {
      connection.set( key, value, 86400, function( err ) {
        if(err) {
          console.error("Error setting value in memcache", err);
          return reject(err);
        } else {
          return resolve();
        }
      });
    });
  });
}

/**
 * Sets a value in memcached with the specified key
 *
 * @param {string} key - The key to map to the value being stored in memcached.
 * @param {object} value - The value to be stored along side the key.
 */
Memcached.prototype.saveKeyValuePairInDatastoreAndSetInMemcached = function( key, value ) {
  const self = this;
  return self.getUnderlyingMemcache()
  .then(( connection ) => {
    return Promise.all([
      self.saveValueToDatastore(key, value),
      self.memcachedSet(key, value)
    ]);
  })
}

/**
 * Removed the key - value pair from memcached that has the specified key
 *
 * @param {string} key - The key to specifiy what value to remove from memcached.
 *
 */
Memcached.prototype.removeFromMemcached = function( key ) {
  const self = this;
  return self.getUnderlyingMemcache()
  .then(( connection ) => {
    return connection.del(key, function( err ) {
      if(err) console.error("Error while trying to remove value from memcahced: ", err);
    })
  });
}

module.exports = Memcached;
