// Copyright 2014 Google Inc. All rights reserved.
//
// Licensed under the Apache License, Version 2.0 (the 'License');
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
// http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an 'AS IS' BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

/*!
 * @module smartcloud/OpenLocation
 */
'use strict';

// Provides an extra precision code, approximately 2x3 meters.
const DEFAULT_PRECISION = 11;

// Code characters to use standard encoding, grid-based encoding used for higher precision
const PAIR_CODE_LENGTH_ = 10;

// A separator used to break the code into two parts to aid memorability.
const SEPARATOR_ = '+';

// The number of characters to place before the separator.
const SEPARATOR_POSITION_ = 8;

// The character used to pad codes.
const PADDING_CHARACTER_ = '0';

// The character set used to encode the values.
const CODE_ALPHABET_ = '23456789CFGHJMPQRVWX';

const ENCODING_BASE_ = CODE_ALPHABET_.length;
const LATITUDE_MAX_ = 90;
const LONGITUDE_MAX_ = 180;
const PAIR_RESOLUTIONS_ = [20.0, 1.0, .05, .0025, .000125];
const GRID_COLUMNS_ = 4;
const GRID_ROWS_ = 5;
const GRID_SIZE_DEGREES_ = 0.000125;
const MIN_TRIMMABLE_CODE_LEN_ = 6;

/**
 * OpenLocation is a collection of functions to help convert geo cordinates to OpenLocation Codes
 *
 * @constructor
 *
 * @example
 * const OpenLocation = OpenLocation Common(options)
 *
 * @param {object} options - Configuration object containing the following properties:
 * @param {number} precision - Precision for open location code (11 default)
 */
function OpenLocation( options ) {
  if (!(this instanceof OpenLocation)) {
    return new OpenLocation(options);
  }
  this.options = options || {};
  this.precision = this.options.precision || DEFAULT_PRECISION;
}

/**
 * Returns the character list used to create codes in this version of Open Location
 */
OpenLocation.prototype.getAlphabet = function() {
  return CODE_ALPHABET_;
};

/**
 * Determines if a code is valid.
 *
 * @param {string} code The string to check.
 * @return {boolean} True if the string is a valid code.
 */
OpenLocation.prototype.isValid = function(code) {
  if(!code || typeof code !== 'string') return false;
  if(code.indexOf(SEPARATOR_) == -1) return false;
  if(code.indexOf(SEPARATOR_) != code.lastIndexOf(SEPARATOR_)) return false;
  if(code.length == 1) return false;
  if(code.indexOf(SEPARATOR_) > SEPARATOR_POSITION_ || code.indexOf(SEPARATOR_) % 2 == 1) return false;

  // We can have an even number of padding characters before the separator,
  // but then it must be the final character.
  if(code.indexOf(PADDING_CHARACTER_) > -1) {
    if(code.indexOf(PADDING_CHARACTER_) == 0) return false;
    // There can only be one group and it must have even length.
    var padMatch = code.match(new RegExp('(' + PADDING_CHARACTER_ + '+)', 'g'));
    if(padMatch.length > 1 || padMatch[0].length % 2 == 1 || padMatch[0].length > SEPARATOR_POSITION_ - 2) {
      return false;
    }
    // If the code is long enough to end with a separator, make sure it does.
    if (code.charAt(code.length - 1) != SEPARATOR_) return false;
  }

  // If there are characters after the separator, make sure there isn't just
  // one of them (not legal).
  if (code.length - code.indexOf(SEPARATOR_) - 1 == 1) return false;

  // Strip the separator and any padding characters.
  code = code.replace(new RegExp('\\' + SEPARATOR_ + '+'), '')
             .replace(new RegExp(PADDING_CHARACTER_ + '+'), '');
  // Check the code contains only valid characters.
  for (var i = 0, len = code.length; i < len; i++) {
    var character = code.charAt(i).toUpperCase();
    if (character != SEPARATOR_ && CODE_ALPHABET_.indexOf(character) == -1) {
      return false;
    }
  }
  return true;
};

/**
 * Determines if a code is a valid short code.
 *
 * @param {string} code The string to check.
 * @return {boolean} True if the string can be produced by removing four or more characters from the start of a valid code.
 */
OpenLocation.prototype.isShort = function( code ) {
  const self = this;
  if(!self.isValid(code)) return false;
  if(code.indexOf(SEPARATOR_) >= 0 && code.indexOf(SEPARATOR_) < SEPARATOR_POSITION_) return true;
  return false;
};

/**
 * Determines if a code is a valid full Open Location Code.
 *
 * @param {string} code The string to check.
 * @return {boolean} True if the code represents a valid latitude and longitude combination.
 */
OpenLocation.prototype.isFull = function(code) {
  const self = this;
  if(!self.isValid(code)) {
    return false;
  }
  // If it's short, it's not full.
  if(self.isShort(code)) {
    return false;
  }

  // Work out what the first latitude character indicates for latitude.
  var firstLatValue = CODE_ALPHABET_.indexOf(code.charAt(0).toUpperCase()) * ENCODING_BASE_;
  if(firstLatValue >= LATITUDE_MAX_ * 2) {
    // The code would decode to a latitude of >= 90 degrees.
    return false;
  }
  if(code.length > 1) {
    // Work out what the first longitude character indicates for longitude.
    var firstLngValue = CODE_ALPHABET_.indexOf(code.charAt(1).toUpperCase()) * ENCODING_BASE_;
    if(firstLngValue >= LONGITUDE_MAX_ * 2) {
      // The code would decode to a longitude of >= 180 degrees.
      return false;
    }
  }
  return true;
};

/**
 * Encode a location into an Open Location Code.
 *
 * @param {number} latitude The latitude in signed decimal degrees
 * @param {number} longitude The longitude in signed decimal degrees
 * @return {string} The code.
 * @throws {Exception} if any of the input values are not numbers.
 */
OpenLocation.prototype.encode = function( latitude, longitude ) {
  const self = this;
  latitude = Number(latitude);
  longitude = Number(longitude);

  if(isNaN(latitude) || isNaN(longitude)) throw new Error('ValueError: Parameters are not numbers');

  latitude = self.clipLatitude(latitude);
  longitude = self.normalizeLongitude(longitude);

  // Latitude 90 needs to be adjusted to be just less, so the returned code can also be decoded.
  if(latitude == 90) latitude = latitude - self.computeLatitudePrecision(self.precision);
  var code = self.encodePairs(latitude, longitude, Math.min(self.precision, PAIR_CODE_LENGTH_));
  if(self.precision > PAIR_CODE_LENGTH_) {
    code += self.encodeGrid(latitude, longitude, self.precision - PAIR_CODE_LENGTH_);
  }
  return code;
};

/**
 * Decodes an Open Location Code into its location coordinates.
 *
 * Returns a CodeArea object that includes the coordinates of the bounding box - the lower left, center and upper right.
 *
 * @param {string} code The code to decode.
 * @return {OpenLocationCode.CodeArea} An object with the coordinates of the area of the code.
 * @throws {Exception} If the code is not valid.
 */
OpenLocation.prototype.decode = function( code ) {
  const self = this;
  if (!self.isFull(code)) {
    throw ('IllegalArgumentException: Passed Open Location Code is not a valid full code: ' + code);
  }

  // scrub string
  code = code.replace(SEPARATOR_, '');
  code = code.replace(new RegExp(PADDING_CHARACTER_ + '+'), '');
  code = code.toUpperCase();

  // Decode the lat/lng pair component.
  var codeArea = self.decodePairs(code.substring(0, PAIR_CODE_LENGTH_));

  if (code.length <= PAIR_CODE_LENGTH_) {
    // no further refinement needed
    return codeArea;
  } else {
    var gridArea = self.decodeGrid(code.substring(PAIR_CODE_LENGTH_));
    return self.CodeArea(
      codeArea.latitudeLo + gridArea.latitudeLo,
      codeArea.longitudeLo + gridArea.longitudeLo,
      codeArea.latitudeLo + gridArea.latitudeHi,
      codeArea.longitudeLo + gridArea.longitudeHi,
      codeArea.codeLength + gridArea.codeLength
    );
  }
};

/**
 * Recover the nearest matching code to a specified location.
 *
 * Given a valid short Open Location Code this recovers the nearest matching
 * full code to the specified location.
 *
 * @param {string} shortCode A valid short code.
 * @param {number} referenceLatitude The latitude to use for the reference location.
 * @param {number} referenceLongitude The longitude to use for the reference location.
 * @return {string} The nearest matching full code to the reference location.
 * @throws {Exception} if the short code is not valid, or the reference position values are not numbers.
 */
OpenLocation.prototype.recoverNearest = function( shortCode, referenceLatitude, referenceLongitude ) {
  const self = this;
  if(!self.isShort(shortCode)) {
    if(self.isFull(shortCode)) {
      return shortCode;
    } else {
      throw 'ValueError: Passed short code is not valid: ' + shortCode;
    }
  }
  referenceLatitude = Number(referenceLatitude);
  referenceLongitude = Number(referenceLongitude);
  if (isNaN(referenceLatitude) || isNaN(referenceLongitude)) {
    throw ('ValueError: Reference position are not numbers');
  }
  // Ensure that latitude and longitude are valid.
  referenceLatitude = self.clipLatitude(referenceLatitude);
  referenceLongitude = self.normalizeLongitude(referenceLongitude);

  // Clean up the passed code.
  shortCode = shortCode.toUpperCase();
  // Compute the number of digits we need to recover.
  var paddingLength = SEPARATOR_POSITION_ - shortCode.indexOf(SEPARATOR_);
  // The resolution (height and width) of the padded area in degrees.
  var resolution = Math.pow(20, 2 - (paddingLength / 2));
  // Distance from the center to an edge (in degrees).
  var halfResolution = resolution / 2.0;

  // Use the reference location to pad the supplied short code and decode it.
  var codeArea = self.decode(self.encode(referenceLatitude, referenceLongitude).substr(0, paddingLength) + shortCode);
  // How many degrees latitude is the code from the reference? If it is more
  // than half the resolution, we need to move it north or south but keep it
  // within -90 to 90 degrees.
  if (referenceLatitude + halfResolution < codeArea.latitudeCenter && codeArea.latitudeCenter - resolution >= -LATITUDE_MAX_) {
    // If the proposed code is more than half a cell north of the reference location,
    // it's too far, and the best match will be one cell south.
    codeArea.latitudeCenter -= resolution;
  } else if (referenceLatitude - halfResolution > codeArea.latitudeCenter && codeArea.latitudeCenter + resolution <= LATITUDE_MAX_) {
    // If the proposed code is more than half a cell south of the reference location,
    // it's too far, and the best match will be one cell north.
    codeArea.latitudeCenter += resolution;
  }

  // How many degrees longitude is the code from the reference?
  if (referenceLongitude + halfResolution < codeArea.longitudeCenter) {
    codeArea.longitudeCenter -= resolution;
  } else if (referenceLongitude - halfResolution > codeArea.longitudeCenter) {
    codeArea.longitudeCenter += resolution;
  }

  return self.encode(codeArea.latitudeCenter, codeArea.longitudeCenter, codeArea.codeLength);
};

/**
 * Remove characters from the start of an OLC code.
 *
 * This uses a reference location to determine how many initial characters
 * can be removed from the OLC code. The number of characters that can be
 * removed depends on the distance between the code center and the reference
 * location.
 *
 * @param {string} code The full code to shorten.
 * @param {number} latitude The latitude to use for the reference location.
 * @param {number} longitude The longitude to use for the reference location.
 * @return {string} The code, shortened as much as possible that it is still the closest matching code to the reference location.
 * @throws {Exception} if the passed code is not a valid full code or the reference location values are not numbers.
 */
OpenLocation.prototype.shorten = function( code, latitude, longitude ) {
  const self = this;
  if(!self.isFull(code)) throw 'ValueError: Passed code is not valid and full: ' + code;
  if (code.indexOf(PADDING_CHARACTER_) != -1) throw 'ValueError: Cannot shorten padded codes: ' + code;
  var code = code.toUpperCase();
  var codeArea = self.decode(code);
  if (codeArea.codeLength < MIN_TRIMMABLE_CODE_LEN_) throw 'ValueError: Code length must be at least ' + MIN_TRIMMABLE_CODE_LEN_;
  // Ensure that latitude and longitude are valid.
  latitude = Number(latitude);
  longitude = Number(longitude);
  if (isNaN(latitude) || isNaN(longitude)) {
    throw ('ValueError: Reference position are not numbers');
  }
  latitude = self.clipLatitude(latitude);
  longitude = self.normalizeLongitude(longitude);
  // How close are the latitude and longitude to the code center.
  var range = Math.max(
      Math.abs(codeArea.latitudeCenter - latitude),
      Math.abs(codeArea.longitudeCenter - longitude));
  for (var i = PAIR_RESOLUTIONS_.length - 2; i >= 1; i--) {
    // Check if we're close enough to shorten. The range must be less than 1/2
    // the resolution to shorten at all, and we want to allow some safety, so
    // use 0.3 instead of 0.5 as a multiplier.
    if (range < (PAIR_RESOLUTIONS_[i] * 0.3)) {
      // Trim it.
      return code.substring((i + 1) * 2);
    }
  }
  return code;
};

/**
 * Clip a latitude into the range -90 to 90.
 *
 * @param {number} latitude
 * @return {number} The latitude value clipped to be in the range.
 */
OpenLocation.prototype.clipLatitude = function( latitude ) {
  return Math.min(90, Math.max(-90, latitude));
};

/**
 * Compute the latitude precision value for a given code length.
 * Lengths <= 10 have the same precision for lat and long, but lengths > 10 have different precisions due to the grid method having fewer columns than rows.
 * @param {number} codeLength
 * @return {number} The latitude precision in degrees.
 */
OpenLocation.prototype.computeLatitudePrecision = function( codeLength ) {
  if (codeLength <= 10) return Math.pow(20, Math.floor(codeLength / -2 + 2));
  return Math.pow(20, -3) / Math.pow(GRID_ROWS_, codeLength - 10);
};

/**
 * Normalize a longitude into the range -180 to 180, not including 180.
 *
 * @param {number} longitude
 * @return {number} Normalized into the range -180 to 180.
 */
OpenLocation.prototype.normalizeLongitude = function(longitude) {
  while (longitude < -180) {
    longitude = longitude + 360;
  }
  while (longitude >= 180) {
    longitude = longitude - 360;
  }
  return longitude;
};

/**
 * Encode a location into a sequence of OLC lat/lng pairs.
 *
 * This uses pairs of characters (longitude and latitude in that order) to represent each step in a 20x20 grid.
 * Each code, therefore, has 1/400th the area of the previous code.
 *
 * This algorithm is used up to 10 digits.
 *
 * @param {number} latitude The location to encode.
 * @param {number} longitude The location to encode.
 * @param {number} codeLength Requested code length.
 * @return {string} The up to 10-digit OLC code for the location.
 */
OpenLocation.prototype.encodePairs = function( latitude, longitude, codeLength ) {
  var code = '';
  // Adjust latitude and longitude so they fall into positive ranges.
  var adjustedLatitude = latitude + LATITUDE_MAX_;
  var adjustedLongitude = longitude + LONGITUDE_MAX_;
  // Count digits - can't use string length because it may include a separator character.
  var digitCount = 0;
  while (digitCount < codeLength) {
    // Provides the value of digits in this place in decimal degrees.
    var placeValue = PAIR_RESOLUTIONS_[Math.floor(digitCount / 2)];
    // Do the latitude - gets the digit for this place and subtracts that for the next digit.
    var digitValue = Math.floor(adjustedLatitude / placeValue);
    adjustedLatitude -= digitValue * placeValue;
    code += CODE_ALPHABET_.charAt(digitValue);
    digitCount += 1;
    // And do the longitude - gets the digit for this place and subtracts that for the next digit.
    digitValue = Math.floor(adjustedLongitude / placeValue);
    adjustedLongitude -= digitValue * placeValue;
    code += CODE_ALPHABET_.charAt(digitValue);
    digitCount += 1;
    // Should we add a separator here?
    if (digitCount == SEPARATOR_POSITION_ && digitCount < codeLength) {
      code += SEPARATOR_;
    }
  }
  if (code.length < SEPARATOR_POSITION_) {
    code = code + Array(SEPARATOR_POSITION_ - code.length + 1).join(PADDING_CHARACTER_);
  }
  if (code.length == SEPARATOR_POSITION_) {
    code = code + SEPARATOR_;
  }
  return code;
};

/**
 * Encode a location using the grid refinement method into an OLC string.
 *
 * The grid refinement method divides the area into a grid of 4x5, and uses a single character to refine the area.
 * This allows default accuracy OLC codes to be refined with just a single character.
 *
 * This algorithm is used for codes longer than 10 digits.
 *
 * @param {number} latitude The location to encode.
 * @param {number} longitude The location to encode.
 * @param {number} codeLength Requested code length.
 * @return {string} The OLC code digits from the 11th digit on.
 */
OpenLocation.prototype.encodeGrid = function( latitude, longitude, codeLength ) {
  var code = '';
  var latPlaceValue = GRID_SIZE_DEGREES_;
  var lngPlaceValue = GRID_SIZE_DEGREES_;
  // Adjust latitude and longitude so they fall into positive ranges and get the offset for the required places.
  latitude += LATITUDE_MAX_;
  longitude += LONGITUDE_MAX_;
  // To avoid problems with floating point, get rid of the degrees.
  latitude = latitude % 1.0;
  longitude = longitude % 1.0;
  var adjustedLatitude = latitude % latPlaceValue;
  var adjustedLongitude = longitude % lngPlaceValue;
  for (var i = 0; i < codeLength; i++) {
    // Work out the row and column.
    var row = Math.floor(adjustedLatitude / (latPlaceValue / GRID_ROWS_));
    var col = Math.floor(adjustedLongitude / (lngPlaceValue / GRID_COLUMNS_));
    latPlaceValue /= GRID_ROWS_;
    lngPlaceValue /= GRID_COLUMNS_;
    adjustedLatitude -= row * latPlaceValue;
    adjustedLongitude -= col * lngPlaceValue;
    code += CODE_ALPHABET_.charAt(row * GRID_COLUMNS_ + col);
  }
  return code;
};

/**
 * Decode an OLC code made up of lat/lng pairs.
 *
 * This decodes an OLC code made up of alternating latitude and longitude characters, encoded using base 20.
 *
 * @param {string} code The code to decode, assumed to be a valid full code, but with the separator removed.
 * @return {OpenLocationCode.CodeArea} The code area object.
 */
OpenLocation.prototype.decodePairs = function( code ) {
  const self = this;
  // Get the latitude and longitude values. These will need correcting from positive ranges.
  var latitude = self.decodePairsSequence(code, 0);
  var longitude = self.decodePairsSequence(code, 1);
  // Correct the values and set them into the CodeArea object.
  return self.CodeArea(
    latitude[0] - LATITUDE_MAX_,
    longitude[0] - LONGITUDE_MAX_,
    latitude[1] - LATITUDE_MAX_,
    longitude[1] - LONGITUDE_MAX_,
    code.length
  );
};

/**
 * Decode either a latitude or longitude sequence.
 *
 * This decodes the latitude or longitude sequence of a lat/lng pair encoding.
 * Starting at the character at position offset, every second character is decoded and the value returned.
 *
 * @param {string} code A valid full OLC code, with the separator removed.
 * @param {string} offset The character to start from.
 * @return {[number]} An array of two numbers, representing the lower and upper range in decimal degrees.
*                     These are in positive ranges and will need to be corrected appropriately.
 */
OpenLocation.prototype.decodePairsSequence = function( code, offset ) {
  var i = 0;
  var value = 0;
  while (i * 2 + offset < code.length) {
    value += CODE_ALPHABET_.indexOf(code.charAt(i * 2 + offset)) * PAIR_RESOLUTIONS_[i];
    i += 1;
  }
  return [value, value + PAIR_RESOLUTIONS_[i - 1]];
};

/**
 * Decode the grid refinement portion of an OLC code.
 *
 * @param {string} code The grid refinement section of a code.
 * @return {OpenLocationCode.CodeArea} The area of the code.
 */
OpenLocation.prototype.decodeGrid = function( code ) {
  const self = this;
  var latitudeLo = 0.0;
  var longitudeLo = 0.0;
  var latPlaceValue = GRID_SIZE_DEGREES_;
  var lngPlaceValue = GRID_SIZE_DEGREES_;
  var i = 0;
  while (i < code.length) {
    var codeIndex = CODE_ALPHABET_.indexOf(code.charAt(i));
    var row = Math.floor(codeIndex / GRID_COLUMNS_);
    var col = codeIndex % GRID_COLUMNS_;

    latPlaceValue /= GRID_ROWS_;
    lngPlaceValue /= GRID_COLUMNS_;

    latitudeLo += row * latPlaceValue;
    longitudeLo += col * lngPlaceValue;
    i += 1;
  }
  return self.CodeArea(
    latitudeLo,
    longitudeLo,
    latitudeLo + latPlaceValue,
    longitudeLo + lngPlaceValue,
    code.length
  );
};

/**
 * Returns an object containing the coordinates include the latitude and longitude of the lower left and
 * upper right corners and the center of the bounding box for the area the code represents.
 *
 * @param {string} latitudeLo Lower left bounding box lat.
 * @param {string} longitudeLo Lower left bounding box long.
 * @param {string} latitudeHi Upper right bounding box lat.
 * @param {string} longitudeHi Upper right bounding box long.
 * @param {string} codeLength Lenght of the open location code.
 * @return {obj} The object containing bounding box coordinates and centers
 */
OpenLocation.prototype.CodeArea = function( latitudeLo, longitudeLo, latitudeHi, longitudeHi, codeLength ) {
  return {
    latitudeLo: latitudeLo,
    longitudeLo: longitudeLo,
    latitudeHi: latitudeHi,
    longitudeHi: longitudeHi,
    codeLength: codeLength,
    latitudeCenter: Math.min(latitudeLo + (latitudeHi - latitudeLo) / 2, LATITUDE_MAX_),
    longitudeCenter: Math.min(longitudeLo + (longitudeHi - longitudeLo) / 2, LONGITUDE_MAX_)
  }
}

module.exports = OpenLocation;
