/*!
 * Copyright 2014 Smart Parking. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*!
 * @module smartcloud/Pubsub
 */

'use strict';
const { PubSub: GCloudPubSub } = require('@google-cloud/pubsub');

/**
 * SmartCloud PubSub is a utility helper for interacting with PubSub for cloud functions
 *
 * @constructor
 *
 * @example
 * const PubSub = new PubSub(options)
 *
 * @param {object} options - Configuration object containing the following properties:
 * @param {string} options.maxRetries - Number of maxRetries
 */
function PubSub( options ) {
  if (!(this instanceof PubSub)) {
    return new PubSub(options);
  }
  this.options = options || {};
  this.maxRetries = this.options.maxRetries || 10;
  this.pubsub;
}

/**
 * Returns registered pubsub instance
 * We instantiate pubsub on first call to prevent current auth error in cloud functions beta
 */
PubSub.prototype.getUnderlyingPubsub = function() {
  const self = this;
  if(!self.pubsub) {
    self.pubsub = new GCloudPubSub({ maxRetries: self.maxRetries });
  }
  return self.pubsub;
};

/**
 * Sends message to defined topic in pubsub
 *
 * @param {string} topic - [Required] Target topic to send message to.
 * @param {object} message - [Required] The object to send.
 *
 */
PubSub.prototype.sendToPubsub = function( topicName, message ) {
  const self = this;
  const topic = self.getUnderlyingPubsub().topic(topicName);

  if(message.constructor === Array) {
    const publishPromises = message.map(( message ) => {
      return topic.publish(Buffer.from(JSON.stringify(message)));
    });
    return Promise.all(publishPromises);
  }

  return topic.publish(Buffer.from(JSON.stringify(message)));
}

/**
 * Loads JSON object from pubsub trigger event
 *
 * @param {object} event - [Required] The event to unwrap.
 *
 */
PubSub.prototype.loadEventData = function( event ) {
  const self = this;
  const pubsubMessage = event.data;
  return JSON.parse(Buffer.from(pubsubMessage.data, 'base64').toString());
}

module.exports = PubSub;
