/*!
 * Copyright 2014 Smart Parking. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*!
 * @module smartcloud/BigTable
 */

'use strict';

/**
 * Result is a wrapper for the rows returned from an hbase query
 *
 * @constructor
 * @param {object} row - decoded JSON string
 */
function Result( row ) {
  if (!(this instanceof Result)) {
    return new Result(row);
  }
  this.row = row || {};
  this.data = this.row.data;
  this.id = this.row.id;
}

/**
 * Returns true if result contains no data
 */
Result.prototype.isEmpty = function( ) {
  if(!this.data) return true;
  return false;
}

/**
 * Returns a value for a specific family:qualifier
 *
 * @constructor
 * @param {string} family - name of the value's family
 * @param {string} qualifier - name of the value's qualifier
 * @param {number} version - which version to return (optional, default: 0 [latest])
 */
Result.prototype.getValue = function( family, qualifier, version ) {
  if(!version) version = 0
  try {
    return this.data[family][qualifier][version].value;
  } catch(error) {
    if (error instanceof TypeError)
      return undefined
    else
      throw error;
  }
}

/**
 * Returns all data as a json object
 *
 * @constructor
 * @param {number} version - which version to return (optional, default: 0 [latest])
 */
Result.prototype.getJson = function( version ) {
  const self = this;
  if(!version) version = 0
  const json = {};
  Object.keys(self.data).forEach(( family ) => {
    json[family] = self.getJsonForFamily(family, version);
  });
  return json;
}

/**
 * Returns values for all qualifiers in family
 *
 * @constructor
 * @param {string} family - name of the family for the qualifiers
 * @param {number} version - which version to return
 */
Result.prototype.getJsonForFamily = function( family, version ) {
  const self = this;
  const json = {};
  Object.keys(self.data[family]).forEach(( qualifier ) => {
    json[qualifier] = self.getValue(family, qualifier, version)
  });
  return json;
}

module.exports = Result;
