/*!
 * Copyright 2014 Smart Parking. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*!
 * @module smartcloud/Setup
 */

'use strict';

const { BigQuery: GCloudBigQuery } = require('@google-cloud/bigquery');
const { Bigtable: GCloudBigTable } = require('@google-cloud/bigtable');
const { PubSub: GCloudPubSub } = require('@google-cloud/pubsub');
const { Storage: GCloudStorage } = require('@google-cloud/storage');
const { Datastore: GCloudDatastore } = require('@google-cloud/datastore');
const fetch = require('node-fetch');
const shelljs = require('shelljs');

const TEST_ENV_VARIABLE_TO_CHECK = 'SMARTCLOUD_PUBSUB_ROUTER_PORT';
const DEFAULT_INSTANCE = 'smart-cloud';
const DEFAULT_REGION = 'us-central1-c';
const DEFAULT_NODES = 3;

/**
 * Setup is a helper for setting up resources needed by Cloud Functions
 *
 * @constructor
 *
 * @example
 * const Setup = new Setup(options)
 *
 * @param {object} options - Configuration object
 */
function Setup( app, options ) {
  if (!(this instanceof Setup)) {
    return new Setup(app, options);
  }
  this.app = app;
  this.options = options || {};
  this.functionName = 'Unknown';
  this.envChecked = false;
  this.testEnvironment = false;
  this.region = this.options.gcpRegion || DEFAULT_REGION;
  this.nodes = this.options.bigtableNodeCount || DEFAULT_NODES;

  this.PubSubModule;
  this.BigQueryModule;
  this.BigTableModule;
  this.DatastoreModule;
  this.StorageModule;
  this.ShellModule;
  this.FetchModule;
}

/**
 * Manually set function being handled by setup
 */
Setup.prototype.setApp = function( app ) {
  this.app = app;
}

/**
 * Manually set options for setup
 */
Setup.prototype.setOptions = function( options ) {
  this.options = options;
}

/**
 * Returns true if a testing environment has been detected
 */
Setup.prototype.isTestingEnvironment = function() {
  return this.testEnvironment;
}

/**
 * Creates formatted log message
 * @private
 */
Setup.prototype.log = function( message ) {
  const self = this;
  console.log(`[${this.functionName}]  `, message)
}

/**
 * Gets variable from env.  Return null if not found.
 */
Setup.prototype.getEnvVariable = function( variableName ) {
  const self = this;
  const value = process.env[variableName];
  if(!value) return undefined;
  if(value == 'undefined') return undefined;
  return value;
};

/**
 * Checks if environment has needed service to perform Setup
 */
Setup.prototype.checkEnvironment = function() {
  const self = this;
  if(self.envChecked) return Promise.resolve();
  return self.checkForTestEnvironment()
  .then(( results ) => { return self.loadAppDetails() })
  .then(( results ) => { return self.checkForGcloudCli() })
  .then(( results ) => { return self.checkEmulatorsRunning() })
  .then(( results ) => {
    self.envChecked = true;
    return Promise.resolve();
  })
};

/**
 * Check if local env variable indicates a testing environment is in use
 * @private
 */
Setup.prototype.checkForTestEnvironment = function() {
  const self = this;
  self.testEnvironment = self.getEnvVariable(TEST_ENV_VARIABLE_TO_CHECK) ? true : false;
  if(self.testEnvironment) {
    console.log("[Testing settings detected in env variables.  Setup will target local emulators.]")
  }
  return Promise.resolve();
}

/**
 * Load details from the associated function
 * @private
 */
Setup.prototype.loadAppDetails = function() {
  const self = this;
  if(self.app) {
    if(self.app.hasOwnProperty('getFunctionDetails')) {
      self.functionName = self.app.getFunctionDetails().functionName;
      self.log(`app config found`)
    } else {
      throw new Error('app does not export getFunctionDetails function.  Please add to continue with setup.');
    }
  }
  return Promise.resolve();
}

/**
 * Checks if we have glcoud installed
 * @private
 */
Setup.prototype.checkForGcloudCli = function() {
  const self = this;
  if(self.Shell().which('gcloud')) {
    return Promise.resolve();
  } else {
    return Promise.reject('glcoud cli is required to run setup.  Refer to: https://cloud.google.com/sdk/downloads');
  }
}

/**
 * Checks if local emulators are running if doing a local deployment
 * @private
 */
Setup.prototype.checkEmulatorsRunning = function() {
  const self = this;
  if(!self.testEnvironment) return Promise.resolve();

  const pubsubEmulatorCheck = new Promise(( resolve, reject ) => {
    const runningProcesses = self.Shell().exec('pgrep -f -c "pubsub-emulator"', {silent:true}).stdout
    if(runningProcesses <= 1) {
      reject('Pubsub emulator not running.  Cannot deploy locally.  Run "gcloud beta emulators pubsub start"');
    } else {
      resolve();
    }
  });

  const bigtableEmulatorCheck = new Promise(( resolve, reject ) => {
    const runningProcesses = self.Shell().exec('pgrep -f -c "bigtable-emulator"', {silent:true}).stdout
    if(runningProcesses <= 1) {
      reject('Bigtable emulator not running.  Cannot deploy locally.  Run "gcloud beta emulators bigtable start"');
    } else {
      resolve();
    }
  });

  const port = self.getEnvVariable('SMARTCLOUD_PUBSUB_ROUTER_PORT');
  const routerHelperCheck = self.Fetch()(`http://localhost:${port}/ping`, { method: "GET" })
  .then(( res ) => { return Promise.resolve(); })
  .catch(( error ) => { return Promise.reject('Pubsub router not running.  Cannot deploy locally.  Run "node functions/auxiliary/test/pubsubRouter.js"')});

  return Promise.all([pubsubEmulatorCheck, bigtableEmulatorCheck, routerHelperCheck]);
}

/**
 * Returns BigQuery instance
 * @private
 */
Setup.prototype.BigQuery = function() {
  const self = this;
  if(!self.BigQueryModule) self.BigQueryModule = new GCloudBigQuery();
  return self.BigQueryModule;
}

/**
 * Returns BigTable instance
 * @private
 */
Setup.prototype.BigTable = function() {
  const self = this;
  if(!self.BigTableModule) self.BigTableModule = new GCloudBigTable();
  return self.BigTableModule;
}

/**
 * Returns PubSub instance
 * @private
 */
Setup.prototype.PubSub = function() {
  const self = this;
  if(!self.PubSubModule) {
    self.PubSubModule = new GCloudPubSub();
  }
  return self.PubSubModule;
}

/**
 * Returns Storage instance
 * @private
 */
Setup.prototype.Storage = function() {
  const self = this;
  if(!self.StorageModule) self.StorageModule = new GCloudStorage();
  return self.StorageModule;
}

/**
 * Returns Datastore instance
 * @private
 */
Setup.prototype.Datastore = function() {
  const self = this;
  if(!self.DatastoreModule) self.DatastoreModule = new GCloudDatastore();
  return self.DatastoreModule;
}

/**
 * Returns Fetch instance
 * @private
 */
Setup.prototype.Fetch = function() {
  const self = this;
  if(!self.FetchModule) self.FetchModule = fetch;
  return self.FetchModule;
}

/**
 * Returns Shell instance
 * @private
 */
Setup.prototype.Shell = function() {
  const self = this;
  if(!self.ShellModule) self.ShellModule = shelljs;
  return self.ShellModule;
}

/**
 * Creates a dataset in BiqQuery
 * @private
 */
Setup.prototype.createDataset = function( datasetName ) {
  const self = this;
  if(self.isTestingEnvironment()) return Promise.resolve();
  return self.BigQuery().createDataset(datasetName)
  .then(( results ) => {
    const dataset = results[0];
    self.log(`> dataset ${dataset.id} created.`);
    return Promise.resolve();
  })
  .catch(( err ) => {
    if(err.message.startsWith('Already Exists: Dataset')) {
      self.log('> dataset already exists');
    } else {
      throw err;
    }
  });
}

/**
 * Creates a table for a dataset
 * @private
 */
Setup.prototype.createBQTable = function( datasetName, tableName, options ) {
  const self = this;
  if(self.isTestingEnvironment()) return Promise.resolve();
  // function updates operations object, so we need to clone it
  return self.BigQuery().dataset(datasetName).createTable(tableName, JSON.parse(JSON.stringify(options)))
  .then(( results ) => {
    const table = results[0];
    self.log(`> Table ${table.id} created.`);
    return Promise.resolve();
  })
  .catch(( err ) => {
    if(err.message.startsWith('Already Exists: Table')) {
      self.log(`> ${tableName} table already exists`);
    } else {
      throw err;
    }
  });
}

/**
 * Updates metadata for table (including creating schema and partitions)
 * @private
 */
Setup.prototype.updateMetadata = function( datasetName, tableName, options ) {
  const self = this;
  if(self.isTestingEnvironment()) return Promise.resolve();
  if(self.options.skipMetaUpdate) return Promise.resolve();
  return self.BigQuery().dataset(datasetName).table(tableName).setMetadata(options)
  .then(( result ) => {
    self.log(`> Schema updated for ${tableName}`);
    return Promise.resolve();
  })
};

Setup.prototype.getBigQueryConfigFromAppConfig = function(extractAction) {
  const self = this;
  const functionDetails = self.app.getFunctionDetails();
  const error = 'function has no BigQuery target.  A dataset name must be supplied when calling createBigQueryAssets().'
  if(!functionDetails.targets) throw new Error(error);

  let configValue;
  try {
    functionDetails.targets.forEach(( target ) => {
      if(target.type.toUpperCase() === 'BIGQUERY' || target.type.toUpperCase() === 'BIGQUERY_PARTITIONED')
      {
        configValue = extractAction(target.options);
        throw new Error('TargetAppConfigFound');
      }
    });
  } catch (err) {
    if (err instanceof Error && err.message === 'TargetAppConfigFound') {
      return configValue;
    }
  }

  throw new Error(error);
};

/**
 * Pulls BigQuery dataset name from app targets
 * @private
 */
Setup.prototype.getBigQueryDatasetNameFromAppConfig = function( ) {
  return this.getBigQueryConfigFromAppConfig((options) => options.datasetName);
};

/**
 * Pulls BigQuery table name from app targets
 * @private
 */
Setup.prototype.getBigQueryTableNameFromAppConfig = function( ) {
  return this.getBigQueryConfigFromAppConfig((options) => options.tableName);
};

/**
 * Creates all assets needed to interact with BigQuery
 */
Setup.prototype.createBigQueryAssets = function(tableName, options ) {
  const self = this;
  let datasetName = '';
  if(!options) options = self.options;
  return self.checkEnvironment()
  .then(( results ) => {
    datasetName = self.getBigQueryDatasetNameFromAppConfig();
    if(!tableName) tableName = self.getBigQueryTableNameFromAppConfig();
    self.log(`creating ${tableName} table in BigQuery`);
    return Promise.resolve();
  })
  .then(( results ) => { return self.createDataset(datasetName) })
  .then(( results ) => { return self.createBQTable(datasetName, tableName, options) })
  .then(( results ) => { return self.updateMetadata(datasetName, tableName, options) })
};

/**
 * Creates as instance in BigTable
 * @private
 */
Setup.prototype.createInstance = function( options ) {
  const self = this;
  if(!options) options = self.options;
  const instanceName = options.bigTableInstance || DEFAULT_INSTANCE;
  const instance = self.BigTable().instance(instanceName);

  const instanceOptions = {
    clusters: [{
      id: `${instanceName}-cluster`,
      name: `${instanceName}-cluster`,
      location: self.region,
      nodes: self.nodes
    }]
  }

  return new Promise(( resolve, reject ) => {
    const callback = function( err, instance, operation ) {
      const errorCheck = function( error ) {
        if(error.message.includes('already exists')) {
          self.log(`> Instance ${instanceName} already exists.`);
          resolve();
        } else {
          reject(error);
        }
      }

      if(err) {
        errorCheck(err);
        return;
      }
      operation
        .on('error', ( error ) => {
          errorCheck(error);
        })
        .on('complete', ( ) => {
          self.log(`> Instance ${instanceName} created.`);
          resolve();
        });
    };
    instance.create(instanceOptions, callback);
  })
}


/**
 * Creates table in BigTable
 * @private
 */
Setup.prototype.createBigTableTable = function( tableName, options ) {
  const self = this;
  if(!options) options = self.options;
  const instanceName = options.bigTableInstance || DEFAULT_INSTANCE;
  const instance = self.BigTable().instance(instanceName);
  const table = instance.table(tableName);

  return table.create()
  .then(( results ) => {
    self.log(`> BigTable ${tableName} table created.`);
    return Promise.resolve();
  })
  .catch(( err ) => {
    if(err.message.includes('already exists')) {
      self.log(`> BigTable ${tableName} table already exists`);
    } else {
      throw err;
    }
  });
}

/**
 * Creates families in BigTable table
 * @private
 */
Setup.prototype.createFamilies = function( tableName, options ) {
  const self = this;
  if(!options) options = self.options;
  if(!options.families) return Promise.resolve();

  const instanceName = options.bigTableInstance || DEFAULT_INSTANCE;
  const instance = self.BigTable().instance(instanceName);
  const table = instance.table(tableName);

  return self.createdFamiliesInSerial(table, options.families );
}

/**
 * Creates families in serial to avoid too-many-operations error
 * @private
 */
Setup.prototype.createdFamiliesInSerial = function( table, families ) {
  const self = this;
  if(families.length == 0) return Promise.resolve();
  const family = families.pop();
  const familyName = family.name || family;
  const familyRule = family.rule || {}
  return table.createFamily(familyName, familyRule)
  .then(( result ) => {
    self.log(`> ${familyName} column family created.`);
    return self.createdFamiliesInSerial( table, families );
  })
  .catch(( err ) => {
    if(err.message.includes('already exists') || err.message.includes('ALREADY_EXISTS')) {
      self.log(`> ${familyName} column family already exists`);
      return self.createdFamiliesInSerial( table, families );
    } else {
      throw err;
    }
  })
}

/**
 * Pulls BigTable table name from app targets
 * @private
 */
Setup.prototype.getBigTableNameFromAppConfig = function( ) {
  const self = this;
  const functionDetails = self.app.getFunctionDetails();
  const error = 'function has no BigTable target.  A table name must be supplied when calling createBigTableAssets().'
  if(!functionDetails.targets) throw new Error(error);
  var tableName;
  functionDetails.targets.forEach(( target ) => {
    if(target.type.toUpperCase() == 'BIGTABLE') {
      tableName = target.name
    }
  });
  if(!tableName) throw new Error(error);
  return tableName;
};

/**
 * Creates all assets needed to interact with BigTable
 */
Setup.prototype.createBigTableAssets = function( tableName, options ) {
  const self = this;
  if(self.isTestingEnvironment()) return Promise.resolve();
  if(!options) options = self.options;
  return self.checkEnvironment()
  .then(( results ) => {
    if(!tableName) tableName = self.getBigTableNameFromAppConfig();
    self.log(`creating ${tableName} table in BigTable`);
    return Promise.resolve();
  })
  .then(( results ) => { return self.createInstance(options) })
  .then(( results ) => { return self.createBigTableTable(tableName, options) })
  .then(( results ) => { return self.createFamilies(tableName, options) })
};

/**
 * Pulls topic name from app targets
 * @private
 */
Setup.prototype.getTopicNameFromAppConfig = function( ) {
  const self = this;
  const functionDetails = self.app.getFunctionDetails();
  const error = 'function has no PubSub target.  A topic name must be supplied when calling createPubSubTopic().'
  if(!functionDetails.targets) throw new Error(error);
  var topicName;
  functionDetails.targets.forEach(( target ) => {
    if(target.type.toUpperCase() == 'PUBSUB') {
      topicName = target.name
    }
  });
  if(!topicName) throw new Error(error);
  return topicName;
};

/**
 * Creates a subscription with local pubsub routing service
 * @private
 */
Setup.prototype.createLocalSubscription = function( functionDetails ) {
  const self = this;
  if(!self.isTestingEnvironment()) return Promise.resolve();
  if(functionDetails.triggerType.toUpperCase() !== 'PUBSUB') return Promise.resolve();

  self.log(`creating local PubSub route`);
  const functionName = functionDetails.functionName;
  const topicName = functionDetails.trigger;

  return self.createTopic(topicName)
  .then(( results ) => { self.createLocalRoute(topicName, functionName) })
}

/**
 * Registers pubsub trigger with local pubsub router
 * @private
 */
Setup.prototype.createLocalRoute = function( topicName, functionName ) {
  const self = this;
  if(!topicName || !functionName) throw new Error('function or trigger name is missing.  Cannot create local subscription.')
  const port = self.getEnvVariable('SMARTCLOUD_PUBSUB_ROUTER_PORT');
  return self.Fetch()(`http://localhost:${port}/pubsub/route/${topicName}/${functionName}`, { method: "POST", body: {} })
  .then(( res ) => {
    if(res.ok) {
      self.log(`> Local topic created for ${topicName}`);
      return Promise.resolve();
    } else {
      return res.text()
      .then(( responseText ) => {
        return Promise.reject(responseText)
      });
    }
  });
}

/**
 * Creates a pubsub topic
 * @private
 */
Setup.prototype.createTopic = function( topicName ) {
  const self = this;
  return self.PubSub().createTopic(topicName)
  .then(( results ) => {
    self.log(`> Topic ${topicName} created`);
    return Promise.resolve();
  })
  .catch(( err ) => {
    if(err.message.includes('already exists')) {
      self.log(`> ${topicName} topic already exists`);
      return Promise.resolve();
    } else {
      throw err;
    }
  });
};

/**
 * Creates assets needed to use PubSub
 */
Setup.prototype.createPubSubTopic = function( topicName, deployedFunctionName ) {
  const self = this;
  return self.checkEnvironment()
  .then(( results ) => {
    if(!topicName) topicName = self.getTopicNameFromAppConfig();
    self.log(`creating ${topicName} topic in pubsub`)
    return Promise.resolve();
  })
  .then(( results ) => { return self.createTopic(topicName) })
};

/**
 * Pulls bucket name from app targets
 * @private
 */
Setup.prototype.getBucketNameFromAppConfig = function( ) {
  const self = this;
  const functionDetails = self.app.getFunctionDetails();
  const error = 'function has no Storage target.  A bucket name must be supplied when calling createStorage().'
  if(!functionDetails.targets) throw new Error(error);
  var bucketName;
  functionDetails.targets.forEach(( target ) => {
    if(target.type.toUpperCase() == 'STORAGE') {
      bucketName = target.name
    }
  });
  if(!bucketName) throw new Error(error);
  return bucketName;
};

/**
 * Creates assets needed to use Google Cloud Storage
 */
Setup.prototype.createStorage = function( bucketName ) {
  const self = this;
  return self.checkEnvironment()
  .then(( results ) => {
    if(!bucketName) bucketName = self.getBucketNameFromAppConfig();
    self.log(`creating ${bucketName} bucket in GCS`);
    return Promise.resolve();
  })
  .then(( results ) => { return self.createStorageBucket(bucketName) })
};

/**
 * Creates a storage bucket
 * @private
 */
Setup.prototype.createStorageBucket = function( bucketName ) {
  const self = this;
  if(self.isTestingEnvironment()) return Promise.resolve();
  return self.Storage().createBucket(bucketName)
  .then(( results ) => {
    self.log(`> Bucket ${bucketName} created`);
    return Promise.resolve();
  })
  .catch(( err ) => {
    if(err.message.startsWith('You already own this bucket')) {
      self.log(`> ${bucketName} bucket already exists`);
      return Promise.resolve();
    } else {
      throw err;
    }
  });
}

/**
 * Output notification for unsupported setup type
 * @private
 */
Setup.prototype.noOp = function( target ) {
  const self = this;
  return self.checkEnvironment()
  .then(( results ) => {
    self.log(`> No registered setup for ${target.type} targets`)
    return Promise.resolve();
  })
}

/**
 * Runs setup for specific target
 * @private
 */
Setup.prototype.setupTarget = function( target ) {
  const self = this;
  switch(target.type.toUpperCase()) {
    case "PUBSUB": return self.createPubSubTopic(target.name);
    case "STORAGE": return self.createStorage();
    case "BIGQUERY": return self.createBigQueryAssets();
    case "BIGQUERY_PARTITIONED": return self.createBigQueryAssets();
    case "BIGTABLE": return self.createBigTableAssets();
    default: return self.noOp(target)
  }
}

/**
 * Run auto setup based on exported function details
 */
Setup.prototype.runSetup = function( ) {
  const self = this;
  const functionDetails = self.app.getFunctionDetails();
  return self.checkEnvironment()
  .then(( results ) => {
    const promises = functionDetails.targets.map(( target ) => self.setupTarget(target));
    return Promise.all(promises);
  })
  .then(( results ) => { return self.createLocalSubscription(functionDetails)})
  .catch(( error ) => {
    self.log(error);
    throw error;
  })
}


module.exports = Setup;
