/*!
 * Copyright 2014 Smart Parking. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*!
 * @module smartcloud/SmartFunction
 */

'use strict';

const MAX_EVENT_AGE = Number.MAX_SAFE_INTEGER; // disable MAX event age by default

// dependencies
const promiseRetry = require('promise-retry');

/**
 * @type {module:smartcloud/smartlog}
 * @private
 */
var SmartLog = require('./smartlog.js');

/**
 * @type {module:smartcloud/errors}
 * @private
 */
var Errors = require('./errors.js');
const logger = require('./winstonLog.js');

/**
 * SmartCloud Funcntion is a wrapper for defining a cloud function to help
 * with logging, messaging, and common workflows
 *
 * @constructor
 *
 * @example
 * const thisFunction = new SmartFunction(options)
 *
 * @param {object} options - Configuration object containing the following properties:
 * @param {string} options.functionName - [Required] Name of this function
 * @param {string} options.trigger - Trigger source for this function required for pubsub and gcs only
 * @param {string} options.triggerType - Type of trigger (e.g., pubsub, gcs, http)
 * @param {string} options.aggregateType - Aggregate type for loading existing records
 * @param {string} options.retry - Adds --retry flag to deployment step when set to true
 * @param {string} options.memory - One of 128MB, 256MB, 512MB, 1024MB, or 2048MB (default 128MB)
 * @param {string} options.timeout - (default 60s by google cloud function)
 * @param {array} options.targets - [Required] Array of targets (type: 'pubsub', name: 'device-commands')
 * @param {array} options.filter - Array of fact subtypes your function wants (e.g. [BayOccupied, BayVacated])
 * @param {int} options.maxRetries - number of max retries when talking to backend services like pubsub (default 10)
 * @param {int} options.maxEventAge - the max time in millisecond that we'll allow an event to be retried (default 10000 // 10 seconds)
 * @param {array} options.indexColumns - column names to index when storing in datastore
 * @param {string} options.storageBucket - set a specific bucket to upload function to during deployment (very optional)
 * @param {boolean} options.skipSourcing - does not extend the source chain (default false), meant for aux system functions
 * @param {boolean} options.skipLogging - skip log output from this library (default false)
 * @param {object} input - Default input for this function (optional; usually used for testing)
 */
function SmartFunction( options ) {
  if (!(this instanceof SmartFunction)) {
    return new SmartFunction(options);
  }

  if(!options.functionName) this.throwFatalError("functionName not defined when creating SmartFunction");
  if(options.triggerType != 'http') {
    if(!options.trigger) this.throwFatalError("trigger not defined when creating SmartFunction");
  }
  if(!options.targets) this.throwFatalError("target not defined when creating SmartFunction");

  if(options.filter) {
    this.allowedSubtypes = [];
    options.filter.forEach(( subtype ) => {
      this.allowedSubtypes[subtype] = subtype;
    });
  }

  if(options.skipLogging) {
    console.log = function(){};
    console.error = function(){};
  }

  this.input = options.input || {};
  this.maxEventAge = options.maxEventAge || MAX_EVENT_AGE;

  this.PubSub;
  this.Datastore;
  this.Memcached;
  this.Config;
  this.BigQuery;
  this.BigTable;
  this.Storage;
  this.Common;

  this.options = options;
}

/**
 * Run function for supplied event and callback
 *
 * @param {promise} operation - The promise to call with decoded input.
 * @param {object} event - The supplied trigger object.
 * @param {function} callback - The trigger callback / response object.
 *
 */
SmartFunction.prototype.run = function( event, callback, operation ) {
  const self = this;
  self.triggerCallback = callback;
  return self.parseEvent(event)
  .then(( input ) => {
    //self.log("Processing started");
    const promises = [operation(input)]
    return Promise.all(promises)
  })
  .then(( results ) => { self.endFunction(results[0]) })
  .catch(( error ) => { self.handleExitError(error) })
}

/**
 * Parses and returns the event based on the funtion's trigger type
 *
 * @param {object} event - The supplied trigger object.
 */
SmartFunction.prototype.parseEvent = function( event ) {
  const self = this;
  switch(self.options.triggerType.toUpperCase()) {
    case "PUBSUB": return self.loadPubsubMessage(event);
    case "HTTP": return self.loadHttpRequest(event);
    case "STORAGE": return self.loadGcsMessage(event);
    default: {
      self.throwFatalError('Trigger type is not yet supported by SmartFunction');
    }
  }
}

/**
 * Checks if event is too old to process due to retries
 * Based on: https://cloud.google.com/functions/docs/writing/retries
 *
 * @param {object} event - The supplied background event
 *
 * @private
 */
SmartFunction.prototype.checkIfEventIsTooOld = function( event ) {
  const self = this;
  const eventAgeMs = Date.now() - Date.parse(event.timestamp);
  if (eventAgeMs > self.maxEventAge) {
    self.abortProcessing(`Dropping event with age[ms]: ${eventAgeMs}`);
  }
}

/**
 * Decode data from a pubsub trigger event
 *
 * @param {object} event - The supplied pubsub message.
 *
 */
SmartFunction.prototype.loadPubsubMessage = function( event ) {
  const self = this;
  const promise = new Promise(( resolve, reject ) => {
    self.checkIfEventIsTooOld(event);
    const decodedMessage = self.pubsub().loadEventData(event);
    self.input = decodedMessage;
    const factSubtype = decodedMessage.fact ? decodedMessage.fact.subtype || decodedMessage.fact.factSubtype : decodedMessage.factSubtype;
    if(!self.options.filter || self.allowedSubtypes[factSubtype]) {
      resolve(decodedMessage);
    } else {
      const message = `${factSubtype} is not supported by this function.  Skipping processing.`;
      self.log(message);
      throw new Errors.FlowControlError(message);
    }
  });
  return promise;
};

/**
 * Decode data from a Google Cloud Storage trigger event
 *
 * @param {object} event - The supplied storage message.
 *
 */
SmartFunction.prototype.loadGcsMessage = function( event ) {
  const self = this;
  const file = event.data;

  if (file.resourceState === 'not_exists') {
    const message = "This is a file deletion.  Will not process.";
    self.log(message);
    throw new Errors.FlowControlError(message);
  }

  if (file.metageneration != 1) {
    const message = "This is a file matadata change.  Will not process.";
    self.log(message);
    throw new Errors.FlowControlError(message);
  }

  return file.get()
  .then(( data ) => {
    var loadedData = data[0];
    try {
      loadedData = JSON.parse(data[0]);
      self.input = loadedData;
    } catch(err) {
      self.error("Object from storage was not JSON.  Returning raw data.  Parsing error: ", err);
      self.input = loadedData;
    }
    return Promise.resolve(loadedData);
  });
};

/**
 * Decode data from a http trigger event
 *
 * @param {object} req - The supplied http request.
 *
 */
SmartFunction.prototype.loadHttpRequest = function( req ) {
  const self = this;
  var incomingData = req.body;
  try {
    incomingData = JSON.parse(req.body);
    self.input = incomingData;
  } catch(err) {
    self.error("Request was not JSON.  Returning raw data.  Parsing error: ", err);
    self.input = incomingData;
  }
  return Promise.resolve(incomingData);
};

/**
 * Sends a pubsub command to update source record trail for a fact
 *
 * @param {string} functionName - The name of the function or service that touched this fact
 * @param {int} timestamp - [optional] The time (in millisecond since the sepoch) the related fact was updated
 * @param {string} factId - [optional] The UUID of the fact being updated, if different from the one found in the input
 *
 * @private
 */
SmartFunction.prototype.updateSource = function( functionName, timestamp, factId ) {
  const self = this;
  if(self.options.skipSourcing && !factId) return Promise.resolve();
  self.factId = factId ? factId : self.input.fact.id;
  if(!self.factId) self.throwFatalError('Fact ID missing.  Cannot update source record');
  const updateCommand = {
    timestamp: timestamp || Date.now(),
    fact: {
      id: self.factId,
      type: 'Command',
      subtype: 'UpdateSource'
    },
    source: {
      name: functionName
    }
  }
  return self.pubsub().sendToPubsub('source-commands', updateCommand);
}

/**
 * Sends object to all specified function targets
 *
 * @param {object} obj - Object to send.
 * @param {string} id - ID for object to save or explicity
 * @param {...*} path - Ancestor heirarhcy and/or file path
 *
 * @example
 * sendOutput(deviceRecord, 'badger.123').then();
 * sendOutput(deviceRecord, 'badger.123', ['org', 'acme', 'site', 'rocketMall']).then();
 */
SmartFunction.prototype.sendOutput = function( obj, id, path ) {
  const self = this;
  return Promise.all(
    self.options.targets.map(( target ) => { 
      return promiseRetry(function (retry, number) {
        if (number > 1) {
          console.log(`attempt sendToTarget to ${(target.type.toLowerCase() === 'bigquery' ? `${target.options.datasetName}.${target.options.tableName}_${id}` : target.name)}: ${number}`);
        }
        return self.sendToTarget(target, obj, id, path)
        .catch((err) => {
          console.error(`SmartFunction sendOutput method hits error for the target - ${JSON.stringify(target)}.  Error details => ${err}`);
          retry(err);
        })
      }, {retries: 3, minTimeout: 500});
    })
  )
}

/**
 * Sends object to output target
 *
 * @param {object} target - Definition of what to send data to
 * @param {object} obj - Object to send to target
 * @param {...*} id - ID for object to save or explicity array path for entity
 *
 * @example
 * sentToTarget({'type':'pubsub', 'name':'device-commands'}, deviceRecord, 'badger.123').then();
 */
SmartFunction.prototype.sendToTarget = function( target, obj, id, path ) {
  const self = this;
  switch(target.type.toUpperCase()) {
    case "PUBSUB": return self.pubsub().sendToPubsub(target.name, obj);
    case "DATASTORE": return self.saveRecordToDatastoreTarget(target.name, obj, id, path);
    case "STORAGE": return self.storage().saveToGCS(obj, target.name, id);
    case "BIGQUERY": return self.bigquery().insertIntoBigQuery(obj, id, target.options);
    case "BIGQUERY_PARTITIONED": return self.bigquery().insertIntoPartitionedBigQuery(obj, target.options);
    case "BIGTABLE": return self.bigtable().insertRowIntoBigTable(obj, id, target.name);
    case "MEMCACHE": return self.saveLabelToMemcache(obj);
    default: {
      self.log(`${target} output for ${id}`, obj);
      return Promise.resolve();
    }
  }
}

/**
 * Sends an error message
 *
 * @param {object} message - error message to send
 *
 */
SmartFunction.prototype.error = function( message ) {
  const self = this;
  console.error(SmartLog(self.options.functionName, self.input, message, true));
}

/**
 * Generate an OperationError object
 *
 * @param {object} message - error message to send
 *
 */
SmartFunction.prototype.OperationError = function( message ) {
  return new Errors.OperationError(message);
}

/**
 * Handle error found within SmartFunction promise chain
 *
 * @param {object} error - the caught error
 *
 */
SmartFunction.prototype.handleExitError = function( error ) {
  const self = this;
  if(error instanceof Errors.FlowControlError) {
    return self.resolveTriggerCallback();
  } else if (error instanceof Errors.FunctionRetryError) {
    return self.rejectTriggerCallback(error.message);
  } else if (error instanceof Errors.NoRetryError || error instanceof Errors.OperationError) {
    return self.resolveTriggerCallback(error);
  }

  if(error instanceof Error) {
    console.error(error);
  } else {
    console.error(new Error(error));
  }
  return self.resolveTriggerCallback();
}

/**
 * Break out of promise chain but error will not trigger a retry
 *
 * @param {string} message - an explination failed permanently
 *
 */
SmartFunction.prototype.throwFatalError = function( message ) {
  throw new Errors.NoRetryError(message);
}

/**
 * Break out of promise chain and end execution without triggering an error
 *
 * @param {string} message - an explination what processing was aborted
 *
 */
SmartFunction.prototype.abortProcessing = function( message ) {
  console.log("Processing aborted: ", message);
  throw new Errors.FlowControlError(message);
}

/**
 * Break out of promise chain and trigger error to retry the function
 */
SmartFunction.prototype.retryProcessing = function ( error ) {
  const self = this;
  if(self.options.retry) {
    if(error instanceof Error) {
      throw new Errors.FunctionRetryError(error.message);
    } else {
      throw new Errors.FunctionRetryError(error);
    }
  } else {
    console.error('Unable retry function - retries for this function are disabled.')
    throw new Errors.FlowControlError()
  }
}

/**
 * Perform any final actions required when a function exits
 *
 * @private
 */
SmartFunction.prototype.endFunction = function( returnValue ) {
  const self = this;
  return self.resolveTriggerCallback(returnValue);
}

/**
 * Call or update the registered callback for this function
 *
 * @private
 */
SmartFunction.prototype.resolveTriggerCallback = function( returnValue ) {
  const self = this;
  switch(self.options.triggerType.toUpperCase()) {
    case "HTTP": return self.triggerCallback.status(200).send(returnValue);
    default: self.triggerCallback();
  }
}

/**
 * Reject or update the registered callback for this function
 *
 * @private
 */
SmartFunction.prototype.rejectTriggerCallback = function( error ) {
  const self = this;
  self.error(error);
  switch(self.options.triggerType.toUpperCase()) {
    case "HTTP": self.triggerCallback.status(500).send();
    default: self.triggerCallback(error);
  }
}

/**
 * Sends a log message
 *
 * @param {object} message - message to send
 *
 */
SmartFunction.prototype.log = function( message, obj ) {
  const self = this;
  const logMessage = SmartLog(self.options.functionName, (obj || self.input), message);
  logger.info(logMessage);
}

/**
 * Sends a log message
 *
 * @param {object} message - message to send
 *
 */
SmartFunction.prototype.logger = function( message ) {
  if (this.input) {
    const hasObjectProperty = (object) => {
      for (const property in object) {
        if (typeof object[property] === 'object' && object[property] !== null) {
          return true;
        }
      }
      return false;
    }

    const flattenedInput  = Object.assign(
      {},
      ...function _flatten(o) {
        return [].concat(...Object.keys(o)
          .map(k =>
            typeof o[k] === 'object' && hasObjectProperty(o[k]) ?
              _flatten(o[k]) :
              ({[k]: o[k]})
          )
        );
      }(this.input)
    );

    if (flattenedInput.fact && flattenedInput.fact.id) {
      logger.info(message, { labels: { id: flattenedInput.fact.id } });
    } else {
      logger.info(message);
    }
  } else {
    logger.info(message);
  }
}

/**
 * Save object to datastore target
 *
 * @param {string} kind - The entity kind of this target (i.e., table name)
 * @param {object} obj - Object to send to save
 * @param {string} id - ID for object to save or explicity array path for entity
 * @param {array} path - [Optional] Ancestor path of saved object
 *
 * @example
 * saveRecordToDatastoreTarget('Devices', deviceRecord, 'badger.123').then();
 */
SmartFunction.prototype.saveRecordToDatastoreTarget = function( kind, obj, id, path ) {
  const self = this;
  var entityPath = []
  if(path) {
    entityPath = path.slice(0, path.length); //make a copy so we don't pollute the original array
    entityPath.push(kind);
    entityPath.push(id);
  } else {
    entityPath = [kind, id]
  }
  return self.datastore().saveRecord(obj, entityPath);
}

/**
 * Save label to memcached
 *
 * @param {object} obj - The obj containing the following properties:
 * @param {string} aggregateId - The key for memcached
 * @param {string} label - The value the memcached
 * @param {string} language - [Optional] Three-letter language code for the tag (defaults to ENG)
 *
 * @example
 * saveLabelToMemcache('Devices', deviceRecord, 'badger.123').then();
 */
SmartFunction.prototype.saveLabelToMemcache = function( obj ) {
  const self = this;
  const key = obj.key || obj.aggregateId;
  if(!key || !obj.label) self.throwFatalError('key/aggregateId or label is missing, cannot update mamcache.');
  var labelLanguage = obj.language || 'ENG';
  var labelKey = `${key.toUpperCase().trim()}.${labelLanguage.toUpperCase().trim()}`;
  return self.memcached().saveKeyValuePairInDatastoreAndSetInMemcached(labelKey, obj.label);
}

/**
 * Load previous record
 *
 * @param {...*} - ID for object to load or explicity array path for entity
 *
 * @example
 * loadPreviousRecord('badger.123').then();
 * loadPreviousRecord(['org', 'acme', 'site', 'rocketMall']).then();
 */
SmartFunction.prototype.loadPreviousRecord = function( path, additionalOptions ) {
  const self = this;
  if(self.options.aggregateSource && self.options.aggregateSource.toUpperCase() == 'BIGTABLE') {
    return self.loadPreviousRecordFromBigTable(path, additionalOptions);
  } else {
    return self.loadPreviousRecordFromDatastore(path, additionalOptions);
  }
}

/**
 * Load previous record from datastore
 *
 * @param {...*} - ID for object to load or explicity array path for entity
 * @private
 */
SmartFunction.prototype.loadPreviousRecordFromDatastore = function( path, additionalOptions ) {
  const self = this;
  return self.loadPreviousRecordsFromDatastore(path, additionalOptions)
  .then(( records ) => {
    return Promise.resolve([records[0], self.input])
  });
}

/**
 * Load previous records from datastore
 *
 * @param {...*} - ID for object to load or explicity array path for entity
 * @private
 */
SmartFunction.prototype.loadPreviousRecordsFromDatastore = function( path, additionalOptions ) {
  const self = this;
  if(!(path instanceof Array)) {
    path = [self.options.aggregateType, path]
  }
  const entityKey = self.datastore().datastoreKey(path);
  return self.datastore().loadFromDatastore(entityKey)
  .then(( [records] ) => {
    return Promise.resolve([records, self.input])
  });
}

/**
 * Load previous record from BigTable
 *
 * @param {string} - Rowkey for object to load
 * @private
 */
SmartFunction.prototype.loadPreviousRecordFromBigTable = function( rowkey, additionalOptions ) {
  const self = this;
  return self.bigtable().getRow(rowkey, self.options.aggregateType.toLowerCase(), additionalOptions)
  .then(( row ) => {
    return Promise.resolve([row, self.input]);
  });
}

/**
 * Update the operation associated with the input
 *
 * @param {string} status - the status to update this operation to
 * @param {string} message - the message to send with this status update
 *
 */
SmartFunction.prototype.updateOperation = function( status, message ) {
  const self = this;
  if(self.options.skipSourcing) return Promise.resolve();
  if(self.input.fact && self.input.fact.origin) {
    if(message instanceof Error || status.toLowerCase() == 'error') {
      if(message instanceof Errors.OperationError) {
        message = message.message;
      } else {
        message = `Processing error, contact SmartCloud support with operation ID: ${self.input.fact.id}`;
      }
    }
    const updateCommand = {
      fact: {
        id: self.input.fact.id,
        type: 'Command',
        subtype: 'UpdateOperation'
      },
      operation: {
        status: status,
        message: message,
        source: self.options.functionName,
        origin: self.input.fact.origin
      },
      timestamp: new Date().getTime(),
    }
    return self.pubsub().sendToPubsub('operation-commands', updateCommand)
  } else {
    return Promise.resolve();
  }
}

// interfaces
SmartFunction.prototype.pubsub = function() {
  const self = this;
  if(!self.PubSub) self.PubSub = require('./pubsub.js')(self.options);
  return self.PubSub;
};

SmartFunction.prototype.datastore = function() {
  const self = this;
  if(!self.Datastore) self.Datastore = require('./datastore.js')(self.options);
  return self.Datastore;
};

SmartFunction.prototype.memcached = function() {
  const self = this;
  if(!self.Memcached) self.Memcached = require('./memcached.js')(self.options);
  return self.Memcached;
};

SmartFunction.prototype.config = function() {
  const self = this;
  if(!self.Config) self.Config = require('./config.js')(self.options);
  return self.Config;
};

SmartFunction.prototype.bigquery = function() {
  const self = this;
  if(!self.BigQuery) self.BigQuery = require('./bigquery.js')(self.options);
  return self.BigQuery;
};

SmartFunction.prototype.storage = function() {
  const self = this;
  if(!self.Storage) self.Storage = require('./storage.js')(self.options);
  return self.Storage;
};

SmartFunction.prototype.common = function() {
  const self = this;
  if(!self.Common) self.Common = require('./common.js')(self.options);
  return self.Common;
};

SmartFunction.prototype.bigtable = function() {
  const self = this;
  if(!self.BigTable) self.BigTable = require('./bigtable.js')(self.options);
  return self.BigTable;
};

// common wrapper
SmartFunction.prototype.hash = function( stringToHash ) { return this.common().hash(stringToHash); };
SmartFunction.prototype.getUUID = function() { return this.common().getUUID(); };
SmartFunction.prototype.getReference = function( obj ) { return this.common().getReference(obj); };

module.exports = SmartFunction;
