/*!
 * Copyright 2014 Smart Parking. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*!
 * @module smartcloud/Storage
 */

'use strict';

const { Storage: GCloudStorage } = require('@google-cloud/storage');

/**
 * SmartCloud Storage is a utility helper for interacting with Google Cloud Storage for cloud functions
 *
 * @constructor
 *
 * @example
 * const Storage = new Storage(options)
 *
 * @param {object} options - Configuration object containing the following properties:
 */
function Storage( options ) {
  if (!(this instanceof Storage)) {
    return new Storage(options);
  }
  this.options = options || {};
  this.storage;
}

/**
 * Returns registered GCS instance
 * We instantiate storage on first call to prevent current auth error in cloud functions beta
 */
Storage.prototype.getUnderlyingStorage = function() {
  const self = this;
  if(!self.storage) {
    self.storage = new GCloudStorage();

    // TODO: watch out the following which comes out recently as of 12 Mar 2018 as the fix may evolve over time
    //     https://github.com/googleapis/nodejs-storage/issues/101
    self.storage.interceptors.push({
      request: reqOpts => {
        reqOpts.forever = true
        return reqOpts
      }
    })
  }
  return self.storage;
};

/**
 * Performs write operation
 * @private
 */
Storage.prototype.writeToGCS = function( obj, bucketName, fileName, options ) {
  const self = this;
  const bucket = self.getUnderlyingStorage().bucket(bucketName);
  const file = bucket.file(fileName, options);
  return file.save(Buffer.from(JSON.stringify(obj), 'utf8'));
}

/**
 * Saves an object to Google Cloud Storage.  Will not upload duplicate files.
 *
 * @param {object} obj - obj to save
 * @param {string} bucketName - name of the bucket to save to
 * @param {string} fileName - name of the file (including path) to save to target bucket
 *
 */
Storage.prototype.saveToGCS = function( obj, bucketName, fileName ) {
  const self = this;
  const options = {
    generation: 0,  /* only upload if this is a first-time upload, aka generation = 0 */
    metadata: {
      contentType: 'application/json'
    }
  }
  return self.writeToGCS(obj, bucketName, fileName, options)
  .catch(( err ) => {
    if(err.errors && err.errors[0] && err.errors[0].reason === 'conditionNotMet') {
      console.warn(`File previously uploaded.  Will not upload again. OBJECT => ${JSON.stringify(obj)}`);
      return Promise.resolve();
    }

    if (err instanceof Error) {
      console.error(err);
    } else {
      console.error(new Error(`OBJECT => ${JSON.stringify(obj)}, ERROR => ${JSON.stringify(err)}`));
    }
  });
}

/**
 * Saves or overwrites a Google Cloud Storage object
 *
 * @param {object} obj - obj to save
 * @param {string} bucketName - name of the bucket to save to
 * @param {string} fileName - name of the file (including path) to save to target bucket
 *
 */
Storage.prototype.upsertToGCS = function( obj, bucketName, fileName ) {
  const self = this;
  const options = {
    metadata: {
      contentType: 'application/json'
    }
  }
  return self.writeToGCS(obj, bucketName, fileName, options);
}


module.exports = Storage;
